import unittest
import warnings

import numpy as np
import pandas as pd

import rubin_scheduler.scheduler.basis_functions as basis_functions
import rubin_scheduler.scheduler.surveys as surveys
from rubin_scheduler.scheduler.example import simple_greedy_survey
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.schedulers import CoreScheduler
from rubin_scheduler.scheduler.utils import ObservationArray


class TestCoreSched(unittest.TestCase):
    def testsched(self):
        # Just set up a very simple survey, one band
        survey = simple_greedy_survey(bandname="r")

        scheduler = CoreScheduler([survey])

        observatory = ModelObservatory()

        # Check that we can update conditions
        scheduler.update_conditions(observatory.return_conditions())

        # Check that we can get an observation out
        obs = scheduler.request_observation()
        assert obs is not None

        # check that we can pull the whole queue
        obs_list = scheduler.request_observation(whole_queue=True)
        assert obs_list is not None
        assert len(scheduler.queue) == 0

        # Check that we can flush the Queue
        scheduler.flush_queue()
        assert len(scheduler.queue) == 0

        # Check that we can add an observation
        scheduler.add_observation(obs)

        # Check dunder methods
        self.assertIsInstance(repr(scheduler), str)
        self.assertIsInstance(str(scheduler), str)

        # Check access methods
        _ = scheduler.get_basis_functions([0, 0])
        _ = scheduler.get_healpix_maps([0, 0])

        # Check survey access methods
        reward_df = scheduler.make_reward_df(observatory.return_conditions())
        self.assertIsInstance(reward_df, pd.DataFrame)
        reward_df = scheduler.make_reward_df(observatory.return_conditions(), accum=False)
        self.assertIsInstance(reward_df, pd.DataFrame)

        obs = scheduler.request_observation()
        surveys_df = scheduler.surveys_df(0)
        self.assertIsInstance(surveys_df, pd.DataFrame)

    def test_record_rewards(self):
        # Create two surveys, so we can have one of the with a basis function
        # with all nans but stiff have the scheduler find observations it can
        # return.
        surveys = [simple_greedy_survey(bandname="r"), simple_greedy_survey(bandname="i")]

        # Modify the 1st survey so that it has an all-nan array basis function.
        new_basis_function = basis_functions.SimpleArrayBasisFunction(np.nan)
        surveys[0].basis_functions.append(new_basis_function)
        surveys[0].basis_weights.append(1)

        scheduler = CoreScheduler(surveys, keep_rewards=True)
        observatory = ModelObservatory()
        scheduler.update_conditions(observatory.return_conditions())

        with warnings.catch_warnings(record=True) as caught_warnings:
            scheduler.request_observation()
            for caught_warning in caught_warnings:
                if issubclass(caught_warning.category, RuntimeWarning):
                    assert str(caught_warning.message) != "All-NaN slice encountered"

        self.assertIsInstance(scheduler.queue_reward_df, pd.DataFrame)

    def test_add_obs(self):
        """Test that add_observation works with slice or array"""

        bfs = [basis_functions.ForceDelayBasisFunction(days_delay=3.0, scheduler_note="survey")]
        survey = surveys.GreedySurvey(bfs, [0.0])
        scheduler = CoreScheduler([survey])

        obs = ObservationArray()
        obs["scheduler_note"] = "survey"
        obs["mjd"] = 100

        scheduler.add_observation(obs)

        recorded = scheduler.survey_lists[0][0].basis_functions[0].survey_features["last_obs_self"].feature

        assert recorded["mjd"] == 100
        assert recorded["scheduler_note"] == "survey"

        bfs = [basis_functions.ForceDelayBasisFunction(days_delay=3.0, scheduler_note="survey")]
        survey = surveys.GreedySurvey(bfs, [0.0])
        scheduler = CoreScheduler([survey])

        # Again, but now add just the row rather than full array
        scheduler.add_observation(obs[0])
        recorded = scheduler.survey_lists[0][0].basis_functions[0].survey_features["last_obs_self"].feature

        assert recorded["mjd"] == 100
        assert recorded["scheduler_note"] == "survey"


if __name__ == "__main__":
    unittest.main()



import os
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime, timedelta
from functools import cache, cached_property
from math import log, sqrt
from typing import List
from warnings import warn

import healpy as hp
import numpy as np
import pandas as pd
import requests

# import rubin_sim
from astropy.coordinates import angular_separation
from astropy.coordinates.earth import EarthLocation
from astropy.time import Time

from rubin_scheduler.scheduler.utils import SchemaConverter
from rubin_scheduler.site_models import Almanac, SeeingModel
from rubin_scheduler.utils import (
    SURVEY_START_MJD,
    Site,
    SysEngVals,
    _angular_separation,
    _approx_altaz2pa,
    pseudo_parallactic_angle,
    rotation_converter,
)

GAUSSIAN_FWHM_OVER_SIGMA: float = 2.0 * sqrt(2.0 * log(2.0))
KNOWN_INSTRUMENTS: List[str] = ["lsstcam", "lsstcomcam", "lsstcomcamsim", "latiss"]
DEFAULT_EXTINCTION_COEFFICIENTS: dict[str, float] = {
    "u": -0.4582,
    "g": -0.2079,
    "r": -0.1223,
    "i": -0.0739,
    "z": -0.0574,
    "y": -0.0947,
}

try:
    import lsst.rsp

    @cache
    def _get_auth(*args, **kwargs) -> tuple[str, str]:
        return ("user", lsst.rsp.get_access_token(*args, **kwargs))

except ImportError:

    @cache
    def _get_auth(token_file: str | None = None) -> tuple[str, str]:
        token: str | None = None

        if token_file is not None:
            with open(token_file, "r") as token_io:
                token = token_io.read()
        elif "ACCESS_TOKEN" in os.environ:
            token = os.environ.get("ACCESS_TOKEN")

        if token is None:
            raise ValueError("No access token found.")

        return ("user", token)


def _guess_consdb_endpoint():
    location = os.getenv("EXTERNAL_INSTANCE_URL", "")
    hostname = os.getenv("HOSTNAME", "")
    if "summit-lsp" in location or hostname == "htcondor.cp.lsst.org":
        endpoint = "https://summit-lsp.lsst.codes/consdb/query"
    else:
        endpoint = "https://usdf-rsp.slac.stanford.edu/consdb/query"

    return endpoint


def query_consdb(
    query: str,
    url: str | None = None,
    token_file: str | None = None,
):
    """Query the consdb

    Parameters
    ----------
    query : `str`
        The SQL query to send to the consdb.
    url : `str`, optional
        The ConsDB REST API URL, or ``None`` to guess.
        Defaults to ``None``.
    token_file : `str` or `None`
        The file from which to read the access token.
        ``None`` for default.

    Returns
    -------
    result : `pandas.DataFrame`
        The result of the query.
    """
    if url is None:
        url = _guess_consdb_endpoint()
    assert url is not None

    auth = _get_auth(token_file)
    params = {"query": query}

    # Requests from the logging urls often fail the 1st time.
    failed_attempts = 0
    response: None | requests.Response = None
    while response is None or response.status_code != 200:
        response = requests.post(url, auth=auth, json=params)
        if response.status_code != 200:
            failed_attempts += 1
            if failed_attempts > 2:
                # If we fail too many times, raise an exception
                # appropriate to the last failure.
                response.raise_for_status()

    messages = response.json()
    query_results = pd.DataFrame(messages["data"], columns=messages["columns"])

    return query_results


# Different instruments (and different numbers of snaps per visit) do or
# will have differenc schema in the consdb, so make a base class with
# expected common functionality, and use derived classes for
# instrument-specific elements.

# Separate different pieces of metadata (columns in a table of visits) into
# different methods to make it easy to individually override them by
# instrument.


# cached_properties are a way to compute values when and only when needed
# but these won't update if the date is changed. In my expected use, that's
# fine; the user should just instantiate another ConsDBVisits instance if
# they want a different database or date. But, what if someone does?
# To avoid this, make it a frozen dataclass, which prevents members from
# being updated.
@dataclass(frozen=True)
class ConsDBVisits(ABC):
    """Visits returned from a query to the consdb.

    Parameters
    ----------
    day_obs : `str`
        The date for which to query the database.
    url: `str` or `None`
        The ConsDB REST API URL, or ``None`` to guess.
        Defaults to ``None``.
    num_nights : `int`
        The number of nights (up to and including day_obs)
        for which to get visits. Defaults to 1.
    token_file : `str` or `None`
        File from which to load the token. If `None`, use the ``lsst.rsp``
        (if available) or the ``ACCESS_TOKEN`` environment variable.
        Defaults to `None`.
    constraints : `str`
        Additional constraints in the SQL query to send to consdb.
    """

    day_obs: str | int
    url: str | None = None
    num_nights: int = 1
    token_file: str | None = None
    constraints: str = ""

    def _have_numeric_values(self, column) -> bool:
        if column not in self.consdb_visits.columns:
            warn(f"Consdb does not have column {column}, guessing values instead")
            return False

        # Called to see if we need to guess values because the
        # consdb doesn't yet fill in what we need
        value_dtype = self.consdb_visits[column].dtype

        # Needed to make mypy happy
        assert isinstance(value_dtype, np.dtype)

        have_values = np.issubdtype(value_dtype, np.number)
        if not have_values:
            warn(f"Consdb does not have values for {column}, guessing values instead")

        return np.issubdtype(value_dtype, np.number)

    @property
    @abstractmethod
    def instrument(self) -> str:
        """The instrument.

        Returns
        -------
        instrument : `str`
            The instrument.
        """
        # Subclasses should return a string constant specifying the relevant
        # instrument.
        # Specified as an property rather than an attribute so it can be
        # abstract.
        raise NotImplementedError

    @property
    @abstractmethod
    def num_exposures(self) -> int:
        """The number of snaps per visit.

        Returns
        -------
        num_exposures : `int`
            The number of snaps per visit.
        """
        # Subclasses should return a string constant specifying the relevant
        # instrument.
        # Specified as an property rather than an attribute so it can be
        # abstract.
        raise NotImplementedError

    @property
    @abstractmethod
    def telescope(self) -> str:
        """The telescope.

        Returns
        -------
        telescope : `str`
            The telescope.
        """
        # Subclasses should return a constant specifying the relevant
        # instrument.
        # Specified as an property rather than an attribute so it can be
        # abstract.
        raise NotImplementedError

    @property
    def pixel_scale(self) -> pd.Series:
        """The pixel scale in asec/pixel.

        Returns
        -------
        pixel_scale : `pd.Series`
            The pixel scale.
        """
        return self.consdb_visits["pixel_scale"]

    @property
    @abstractmethod
    def site(self) -> Site:
        """The site.

        Returns
        -------
        site : `rubin_scheduler.utils.site.Site`
            The observatory site.
        """
        # Subclasses should return a constant specifying the relevant
        # instrument.
        # Specified as an property rather than an attribute so it can be
        # abstract.
        raise NotImplementedError

    @cached_property
    def almanac(self) -> Almanac:
        """The astronomical almanac.

        Returns
        -------
        almanac : `rubin_scheduler.site_models.almanac.Almanac`
            Astronomical almanac data.
        """
        return Almanac()

    @cached_property
    def location(self) -> EarthLocation:
        """The observatory location.

        Returns
        -------
        location : `astropy.coordinates.earth.EarthLocation`
            The observatory location.
        """
        #
        return EarthLocation(lat=self.site.latitude, lon=self.site.longitude, height=self.site.height)

    @cached_property
    def day_obs_time(self) -> Time:
        """The day_obs as an astropy time.

        Returns
        -------
        day_obs : `astropy.time.Time`
            The astropy time representation of day_obs.
        """
        time: Time = Time(self.day_obs)
        return time

    @cached_property
    def day_obs_mjd(self) -> int:
        """The day_obs as an mjd.

        Returns
        -------
        day_obs_mjd : `int`
            The mjd representation of day_obs.
        """
        # make mypy happy
        assert isinstance(self.day_obs_time.mjd, float)
        return int(self.day_obs_time.mjd)

    @cached_property
    def day_obs_int(self) -> int:
        """The day_obs as an integer.

        Returns
        -------
        day_obs_int : `int`
            The integer representation of day_obs.
        """
        # make mypy happy
        assert isinstance(self.day_obs_time.iso, str)
        return int(self.day_obs_time.iso[:10].replace("-", ""))

    @cached_property
    def consdb_visits(self) -> pd.DataFrame:
        """The visit data as returned from the consdb

        Returns
        -------
        consdb_visits : `pd.DataFrame`
            The visit data.
        """
        day_obs_date = datetime.strptime(f"{self.day_obs_int}", "%Y%m%d").date()
        prior_day_obs_date = day_obs_date - timedelta(days=self.num_nights)
        prior_day_obs_int = int(prior_day_obs_date.strftime("%Y%m%d"))

        # To avoid duplicate columns, explicitly include list of columns to
        # to return in the query instead of using a "SELECT *".
        def columns_query(table, instrument):
            return f"""
                SELECT column_name FROM information_schema.columns
                WHERE table_name = '{table}'
                  AND table_schema = 'cdb_{instrument}'
            """

        visit_columns = (
            query_consdb(columns_query(f"visit{self.num_exposures}", self.instrument), url=self.url)
            .loc[:, "column_name"]
            .values
        )
        ql_columns = (
            query_consdb(columns_query(f"visit{self.num_exposures}_quicklook", self.instrument), url=self.url)
            .loc[:, "column_name"]
            .values
        )
        visit_query_visit_columns = ", ".join([f"v.{c} AS {c}" for c in visit_columns])
        visit_query_ql_columns = ", ".join(
            [f"v{self.num_exposures}q.{c} AS {c}" for c in ql_columns if c not in visit_columns]
        )
        visit_query_columns = ", ".join([visit_query_visit_columns, visit_query_ql_columns])

        consdb_visits_query: str = f"""
            SELECT {visit_query_columns}
            FROM cdb_{self.instrument}.visit1 AS v
            LEFT JOIN cdb_{self.instrument}.visit{self.num_exposures}_quicklook
                AS v{self.num_exposures}q ON v.visit_id = v{self.num_exposures}q.visit_id
            WHERE v.obs_start_mjd IS NOT NULL
                AND v.s_ra IS NOT NULL
                AND v.s_dec IS NOT NULL
                AND v.sky_rotation IS NOT NULL
                AND ((v.band IS NOT NULL) OR (v.physical_filter IS NOT NULL))
                AND v.day_obs <= {self.day_obs_int}
                AND v.day_obs > {prior_day_obs_int}
        """
        if len(self.constraints) > 0:
            consdb_visits_query += f" AND ({self.constraints})"

        return query_consdb(consdb_visits_query, self.url, self.token_file)

    @cached_property
    def visit_id(self) -> pd.Series:
        """The unique identifier from the consdb database.

        Returns
        -------
        visit_id : `pd.Series`
            The unique identifier from the consdb database.
        """
        return self.consdb_visits["visit_id"]

    @cached_property
    def observation_id(self) -> pd.Series:
        """The unique identifier from the consdb database.

        Returns
        -------
        observation_id : `pd.Series`
            The unique identifier from the consdb database.
        """
        # Instead of just using visit_id, create a separate property
        # that can be overridden by subclasses which might not get a valid
        # visit_id.
        return self.visit_id

    @cached_property
    def ra(self) -> pd.Series:
        """The R.A. of each visit, in degrees.

        Returns
        -------
        ra : `pd.Series`
            The R.A. of each visit, in degrees.
        """
        return self.consdb_visits["s_ra"]

    @cached_property
    def decl(self) -> pd.Series:
        """The declination if each visit, in degrees.

        Returns
        -------
        decl : `pd.Series`
            The declination of each visit, in degrees.
        """
        return self.consdb_visits["s_dec"]

    @cached_property
    def obs_start_mjd(self) -> pd.Series:
        """The MJD at the start of each visit.

        Returns
        -------
        obs_start_mjd : `pd.Series`
            The MJD at the start of each visit.
        """
        return self.consdb_visits["obs_start_mjd"]

    @cached_property
    def shut_time(self) -> pd.Series:
        """Spatially-averaged shutter-open duration of each visit.

        Returns
        -------
        shut_time : `pd.Series`
            Spatially-averaged shutter-open duration.
        """
        return self.consdb_visits["shut_time"]

    @cached_property
    def obs_midpt_mjd(self) -> pd.Series:
        """Midpoint time for exposure at the fiducial center of the focal
        plane. array in MJD.

        Returns
        -------
        midpt_mjd: `pd.Series`
            Midpoint time for exposure at the fiducial center of the focal
            plane. array in MJD.
        """
        if self._have_numeric_values("obs_midpt_mjd"):
            mjd = self.consdb_visits["obs_midpt_mjd"]
        elif self._have_numeric_values("obs_end_mjd"):
            mjd = (self.obs_start_mjd + self.consdb_visits["obs_end_mjd"]) / 2
        else:
            mjd = self.obs_start_mjd + (0.5 * self.shut_time) / 86400

        return mjd

    @cached_property
    def sky_rotation(self) -> pd.Series:
        """Targeted sky rotation angle.

        Returns
        -------
        sky_rotation : `pd.Series`
            Targeted sky rotation angle for each visit.
        """
        return self.consdb_visits["sky_rotation"]

    @cached_property
    def target_name(self) -> pd.Series:
        """The target name for each visit.

        Returns
        -------
        target_name : `pd.Series`
            The target name for each visit.
        """
        return self.consdb_visits["target_name"]

    @cached_property
    def airmass(self) -> pd.Series:
        """Airmass of each visit.

        Returns
        -------
        airmass : `pd.Series`
            Airmass of each visit.
        """
        return self.consdb_visits["airmass"]

    @cached_property
    def band(self) -> pd.Series:
        """Abstract filter for each visit

        Returns
        -------
        band : `pd.Series`
            Abstract filter for each visit.
        """
        return self.consdb_visits["band"]

    @cached_property
    def extinction_coefficient_without_clouds(self) -> pd.Series:
        """Atomspheric extinction coefficient for each visit, without clouds.

        Returns
        -------
        extinction_coeffcient : `pd.Series`
            Extinction coefficient for each visit.
        """
        # pd.Series.map maps missing keys to nan
        return self.band.map(DEFAULT_EXTINCTION_COEFFICIENTS)

    @cached_property
    def cloud_extinction(self) -> pd.Series:
        """Extinction due to clouds.

        Returns
        -------
        cloud_extinction : `pd.Series`
            The extinction due to clouds
        """
        warn("Measurment of cloud extinction not available; assuming photometric conditions.")
        return pd.Series(0.0, index=self.consdb_visits.index)

    @cached_property
    def atmospheric_extinction_with_clouds_and_air(self) -> pd.Series:
        """The total extinction for each visit.

        Returns
        -------
        extinction : `pd.Series`
            The extinction of each visit, in magnitudes.
        """
        return self.airmass * self.extinction_coefficient_without_clouds + self.cloud_extinction

    @cached_property
    def azimuth(self) -> pd.Series:
        """Azimuth for each visit (degrees) at central time.

        Returns
        -------
        azimuth : `pd.Series`
            Azimuth for each visit.
        """
        return self.consdb_visits["azimuth"]

    @cached_property
    def altitude(self) -> pd.Series:
        """Altitude for each visit (degrees) at central time.

        Returns
        -------
        altitude : `pd.Series`
            Altitude for each visit (degrees).
        """
        return self.consdb_visits["altitude"]

    @cached_property
    def zd(self) -> pd.Series:
        """Zenith distance for each visit at central time (degrees)

        Returns
        -------
        zd : `pd.Series`
            Zenith distance for each visit at central time (degrees)
        """
        return 90 - self.altitude

    @cached_property
    def note(self) -> pd.Series:
        """Note for each visit.

        Returns
        -------
        note : `pd.Series`
            Note for each visit.
        """
        if "note" in self.consdb_visits:
            note = self.consdb_visits["note"]
        else:
            note = pd.Series("", index=self.consdb_visits.index)
        return note

    @cached_property
    def scheduler_note(self) -> pd.Series:
        """scheduler_note for each visit.

        Returns
        -------
        note : `pd.Series`
            Note for each visit.
        """
        if "scheduler_note" in self.consdb_visits:
            note = self.consdb_visits["scheduler_note"]
        else:
            note = pd.Series("", index=self.consdb_visits.index)
        return note

    @cached_property
    def observation_reason(self) -> pd.Series:
        """Observation reason for each visit.

        Returns
        -------
        observation_reason : `pd.Series`
            Observation reason for each visit.
        """
        if "observation_reason" in self.consdb_visits:
            reason = self.consdb_visits["observation_reason"]
        else:
            reason = pd.Series("", index=self.consdb_visits.index)
        return reason

    @cached_property
    def science_program(self) -> pd.Series:
        """Science program each visit.

        Returns
        -------
        science_progarm : `pd.Series`
            Science program for each visit.
        """
        if "science_program" in self.consdb_visits:
            science_program = self.consdb_visits["science_program"]
        else:
            science_program = pd.Series("", index=self.consdb_visits.index)
        return science_program

    @cached_property
    def visit_time(self) -> pd.Series:
        """Visit time for each visit (seconds).

        Returns
        -------
        visit_time : `pd.Series`
            Visit time for each visit.
        """
        return (self.consdb_visits["obs_end_mjd"] - self.consdb_visits["obs_start_mjd"]) * 24 * 60 * 60

    @cached_property
    def cloud(self) -> pd.Series:
        """Fractional cloud cover of the sky for each visit.

        Returns
        -------
        cloud : `pd.Series`
            Fractional cloud cover of the sky for each visit.
        """
        return pd.Series(np.nan, index=self.consdb_visits.index)

    @cached_property
    def slew_distance(self) -> pd.Series:
        """Slew distance to reach each visit (degrees).

        Returns
        -------
        slew_distance : `pd.Series`
            Slew distance to reach each visit (degrees).
        """
        slew_distance = pd.Series(np.nan, index=self.consdb_visits.index)
        previous_visit_idx = None
        for visit_idx in self.consdb_visits.index:
            if previous_visit_idx is not None:
                this_slew_distance = angular_separation(
                    np.radians(self.ra[previous_visit_idx]),
                    np.radians(self.decl[previous_visit_idx]),
                    np.radians(self.ra[visit_idx]),
                    np.radians(self.decl[visit_idx]),
                )
                slew_distance[visit_idx] = np.degrees(this_slew_distance)
            previous_visit_idx = visit_idx
        return slew_distance

    @cached_property
    def start_time(self) -> Time:
        """Start time for each visit.

        Returns
        -------
        start_time : `Time`
            Start time for each visit.
        """
        return Time(self.obs_start_mjd, format="mjd")

    @cached_property
    def observation_start_lst(self) -> pd.Series:
        """Apparent local sidereal time (degrees) at the start of each visit.

        Returns
        -------
        lst : `pd.Series`
            Apparent local sidereal time (degrees) at the start of each visit.
        """
        start_lst = pd.Series(
            self.start_time.sidereal_time("apparent", self.location).degree, index=self.consdb_visits.index
        )
        return start_lst

    @cached_property
    def pseudo_parallactic_angle(self) -> pd.Series:
        """Parallactic angle as defined in SMTN-019 for each visit (degrees).

        Returns
        -------
        pa : `pd.Series`
            Parallactic angle as defined in SMTN-019 for each visit.
        """
        # Following sim_runner
        # Using pseudo_parallactic_angle, see https://smtn-019.lsst.io/v/DM-44258/index.html

        # make mypy happy
        assert isinstance(self.decl.values, np.ndarray)
        assert isinstance(self.ra.values, np.ndarray)
        assert isinstance(self.obs_start_mjd.values, np.ndarray)
        assert isinstance(self.site.longitude, float)
        assert isinstance(self.site.latitude, float)
        assert isinstance(self.site.height, float)

        pa, alt, az = pseudo_parallactic_angle(
            self.ra.values,
            self.decl.values,
            self.obs_start_mjd.values,
            lon=self.site.longitude,
            lat=self.site.latitude,
            height=self.site.height,
        )
        return pa

    @cached_property
    def rot_tel_pos(self) -> pd.Series:
        """Telescope rotator position (degrees).

        Returns
        -------
        rot_tel_pos : pd.Series
            Telescope rotator position (degrees).
        """
        rot_tel_pos = rotation_converter(telescope=self.telescope)._rotskypos2rottelpos(
            np.radians(self.sky_rotation), np.radians(self.pseudo_parallactic_angle)
        )
        return np.degrees(rot_tel_pos)

    @cached_property
    def parallactic_angle(self) -> pd.Series:
        """Parallactic angle (degrees).

        Returns
        -------
        pa : `pd.Series`
            Parallactic angle (degrees).
        """
        # Following sim_runner
        pa = pd.Series(
            np.degrees(
                _approx_altaz2pa(np.radians(self.altitude), np.radians(self.azimuth), self.site.latitude_rad)
            ),
            index=self.consdb_visits.index,
        )
        return pa

    @cached_property
    def night(self) -> int:
        """The night of the survey.

        Returns
        -------
        night : `int`
            The night of the survey.
        """
        night: int = 1 + self.day_obs_mjd - int(np.floor(SURVEY_START_MJD))
        return night

    @cached_property
    def sun_moon_positions(self) -> pd.DataFrame:
        """Sun and moon R.A., Declination, Azimuth, and Altitude (degrees)
        at each visit.

        Returns
        -------
        positions : `pd.DataFrame`
            Sun and moon positions (degrees).
        """
        sun_moon = pd.DataFrame(
            self.almanac.get_sun_moon_positions(self.obs_midpt_mjd.values), index=self.consdb_visits.index
        )

        # Convert to degrees, because everything else is in degrees
        for body in ("sun", "moon"):
            for coord in ("alt", "az", "RA", "dec"):
                column = f"{body}_{coord}"
                sun_moon[column] = np.degrees(sun_moon[column])

        return sun_moon

    @cached_property
    def moon_distance(self) -> pd.Series:
        """Angle between each visit pointing the moon (degrees).

        Returns
        -------
        distance : `pd.Series`
            Angle between each visit pointing the moon (degrees).
        """
        # make mypy happy
        assert isinstance(self.decl.values, np.ndarray)
        assert isinstance(self.ra.values, np.ndarray)

        moon_ra = self.sun_moon_positions["moon_RA"].values
        moon_decl = self.sun_moon_positions["moon_dec"].values
        assert isinstance(moon_ra, np.ndarray)
        assert isinstance(moon_decl, np.ndarray)

        moon_distance_rad = _angular_separation(
            np.radians(self.ra.values),
            np.radians(self.decl.values),
            np.radians(moon_ra),
            np.radians(moon_decl),
        )
        return pd.Series(np.degrees(moon_distance_rad), index=self.consdb_visits.index)

    @cached_property
    def solar_elong(self) -> pd.Series:
        """Solar elongation (degrees) of each visit.

        Returns
        -------
        elong : `pd.Series`
            Solar elongation of each visit in degrees.
        """
        # make mypy happy
        assert isinstance(self.decl.values, np.ndarray)
        assert isinstance(self.ra.values, np.ndarray)

        sun_ra = self.sun_moon_positions["sun_RA"].values
        sun_decl = self.sun_moon_positions["sun_dec"].values
        assert isinstance(sun_ra, np.ndarray)
        assert isinstance(sun_decl, np.ndarray)

        solar_elong_rad = _angular_separation(
            np.radians(self.ra.values),
            np.radians(self.decl.values),
            np.radians(sun_ra),
            np.radians(sun_decl),
        )
        return pd.Series(np.degrees(solar_elong_rad), index=self.consdb_visits.index)

    @cached_property
    @abstractmethod
    def _zero_point_column(self) -> str:
        raise NotImplementedError

    @cached_property
    def zero_point(self) -> pd.Series:
        """Photometric zero point for each visit, with "counts" in electons.

        Returns
        -------
        zp : `pd.Series`
            Photometric zero point for each visit, with "counts" in electons.
        """
        return self.consdb_visits[self._zero_point_column]

    @cached_property
    @abstractmethod
    def _sky_bg_column(self) -> str:
        raise NotImplementedError

    @cached_property
    def sky_e_per_pixel(self) -> pd.Series:
        """Median sky background of each visit (in electons).

        Returns
        -------
        sky : `pd.Series`
            Median sky background of each visit (in electons).
        """
        if self._have_numeric_values(self._sky_bg_column):
            sky = self.consdb_visits[self._sky_bg_column]
        else:
            sky = pd.Series(np.nan, index=self.consdb_visits.index)
        return sky

    @cached_property
    def sky_mag_per_asec(self) -> pd.Series:
        """Sky background in each visit, in mags asec^-2

        Returns
        -------
        sky : `pd.Series`
            Sky background in each visit, in mags asec^-2
        """
        sky_zero_point = self.zero_point - self.atmospheric_extinction_with_clouds_and_air
        sky_mag = sky_zero_point - 2.5 * np.log10(self.sky_e_per_pixel / (self.pixel_scale**2))

        return sky_mag

    @cache
    def hpix(self, nside: int) -> pd.Series:
        """The heaplix of the pointing for each visit, for a given nside.

        Parameters
        ----------
        nside : `int`
            The nside of the healpix map to use.

        Returns
        -------
        hpix : `pd.Series`
            The heaplix of the pointing for each visit, for a given nside.
        """
        return pd.Series(hp.ang2pix(nside, self.ra, self.decl, lonlat=True), index=self.consdb_visits.index)

    @cached_property
    @abstractmethod
    def _psf_sigma_column(self) -> str:
        raise NotImplementedError

    @cached_property
    def fwhm_eff(self) -> pd.Series:
        """Effective PSF FWHM (arcseconds).

        Returns
        -------
        fwhm : `pd.Series`
            Effective PSF FWHM (arcseconds).
        """
        return self.consdb_visits[self._psf_sigma_column] * GAUSSIAN_FWHM_OVER_SIGMA * self.pixel_scale

    @cached_property
    def fwhm_geom(self) -> pd.Series:
        """Geometric PSF FWHM (arcseconds).

        Returns
        -------
        fwhm : `pd.Series`
            Geometric PSF FWHM (arcseconds).
        """
        return SeeingModel.fwhm_eff_to_fwhm_geom(self.fwhm_eff)

    @cached_property
    @abstractmethod
    def _seeing_zenith_500nm_column(self) -> str:
        raise NotImplementedError

    @cached_property
    def fwhm_500(self) -> pd.Series:
        """FWHM at zenith and 500nm, in arcseconds, at the time of each visit.

        Returns
        -------
        fwhm : `pd.Series`
            FWHM at zenith and 500nm, in arcseconds, at the time of each visit.
        """
        return (
            self.consdb_visits[self._seeing_zenith_500nm_column] * GAUSSIAN_FWHM_OVER_SIGMA * self.pixel_scale
        )

    @cached_property
    def eff_time_m5(self) -> pd.Series:
        """Effective exposure time, in seconds.
        This is the time to reach the m5 depth in each visit, under
        a reference set of conditions.

        Returns
        -------
        eff_time : `pd.Series`
            The effective exposure time.
        """
        return self.consdb_visits["eff_time_m5"]

    @cached_property
    def opsim(self) -> pd.DataFrame:
        """The table of visits, in a format replicating opsim output.

        Returns
        -------
        opsim : `pd.DataFrame`
            The table of visits, in a format replicating opsim output.
        """

        opsim_dict = {
            "observationId": self.observation_id,
            "fieldRA": self.ra,
            "fieldDec": self.decl,
            "observationStartMJD": self.obs_start_mjd,
            "flush_by_mjd": np.nan,
            "visitExposureTime": self.shut_time,
            "band": self.band,
            "rotSkyPos": self.sky_rotation,
            "rotSkyPos_desired": np.nan,
            "numExposures": self.num_exposures,
            "airmass": self.airmass,
            "seeingFwhm500": self.fwhm_500,
            "seeingFwhmEff": self.fwhm_eff,
            "seeingFwhmGeom": self.fwhm_geom,
            "skyBrightness": self.sky_mag_per_asec,
            "night": self.night,
            "slewTime": np.nan,
            "visitTime": self.visit_time,
            "slewDistance": self.slew_distance,
            "fiveSigmaDepth": self.eff_time_m5,
            "altitude": self.altitude,
            "azimuth": self.azimuth,
            "paraAngle": self.parallactic_angle,
            "psudoParaAngle": self.pseudo_parallactic_angle,
            "cloud": self.cloud,
            "moonAlt": self.sun_moon_positions["moon_alt"],
            "sunAlt": self.sun_moon_positions["sun_alt"],
            "note": self.note,
            "scheduler_note": self.scheduler_note,
            "target_name": self.target_name,
            #            "block_id": "NOT AVAILABLE",
            "observationStartLST": self.observation_start_lst,
            "rotTelPos": self.rot_tel_pos,
            "rotTelPos_backup": np.nan,
            "moonAz": self.sun_moon_positions["moon_az"],
            "sunAz": self.sun_moon_positions["sun_az"],
            "sunRA": self.sun_moon_positions["sun_RA"],
            "sunDec": self.sun_moon_positions["sun_dec"],
            "moonRA": self.sun_moon_positions["moon_RA"],
            "moonDec": self.sun_moon_positions["moon_dec"],
            "moonDistance": self.moon_distance,
            "solarElong": self.solar_elong,
            "moonPhase": self.sun_moon_positions["moon_phase"],
            "cummTelAz": np.nan,
            #            "scripted_id": "NOT AVAILABLE",
            "observation_reason": self.observation_reason,
            "science_program": self.science_program,
            "start_date": self.consdb_visits["obs_start"],
        }

        opsim_df = pd.DataFrame(opsim_dict)

        return opsim_df

    @cached_property
    def obs(self) -> np.recarray:
        """The table of visits, in a format replicating an observation array.

        Returns
        -------
        obs : `np.recarray`
            The observation array.
        """
        return SchemaConverter().opsimdf2obs(self.opsim)

    @cached_property
    def merged_opsim_consdb(self):
        merged_visits = self.consdb_visits.merge(
            self.opsim, left_on="visit_id", right_on="observationId", suffixes=(None, "_opsim")
        )
        return merged_visits


class PachonConsDBVisits(ConsDBVisits):

    @property
    def site(self):
        return Site("LSST")


class SimonyiConsDBVisits(PachonConsDBVisits):

    @property
    def telescope(self) -> str:
        return "rubin"


class LSSTCamConsDBVisits(SimonyiConsDBVisits):

    # Set "constant" properties using cached_property rather
    # than actual attributes, because abstract "members" can only be declared
    # as such using properties, and overriding a property with an actual
    # attribute confuses type checkers.

    @property
    def instrument(self) -> str:
        return "lsstcam"

    @property
    def num_exposures(self) -> int:
        return 1

    @property
    def pixel_scale(self) -> pd.Series:
        if self._have_numeric_values("pixel_scale"):
            pixel_scale = super().pixel_scale
        else:
            pixel_scale = pd.Series(0.2, index=self.consdb_visits.index)
        return pixel_scale

    @cached_property
    def _zero_point_column(self) -> str:
        return "zero_point_median"

    @cached_property
    def observation_id(self) -> pd.Series:
        """The unique identifier from the consdb database.

        Returns
        -------
        visit_id : `pd.Series`
            The unique identifier from the consdb database.
        """
        if self._have_numeric_values("visit_id"):
            observation_id = self.consdb_visits["visit_id"]
        elif self._have_numeric_values("day_obs") and self._have_numeric_values("seq_num"):
            observation_id = self.consdb_visits["day_obs"] * 10000 + self.consdb_visits["seq_num"]
            warn("Cannot use visit_id as observation_id, guessing instead!")
        else:
            warn("Can use neither visit_id nor guess observation_id, just making something up!")
            observation_id = pd.Series(
                Time(self.consdb_visits.obs_start).strftime("9%Y%m%d%H%M%S").astype(int),
                index=self.consdb_visits.index,
            )

        return observation_id

    @cached_property
    def azimuth(self) -> pd.Series:
        if self._have_numeric_values("azimuth"):
            azimuth = super().azimuth
        elif self._have_numeric_values("azimuth_start"):
            if self._have_numeric_values("azimuth_end"):
                azimuth = (self.consdb_visits["azimuth_start"] + self.consdb_visits["azimuth_end"]) / 2
            else:
                azimuth = self.consdb_visits["azimuth_start"]
        else:
            azimuth = pd.Series(np.nan, index=self.consdb_visits.index)

        return azimuth

    @cached_property
    def altitude(self) -> pd.Series:
        if self._have_numeric_values("altitude"):
            altitude = super().altitude
        elif self._have_numeric_values("altitude_start"):
            if self._have_numeric_values("altitude_end"):
                altitude = (self.consdb_visits["altitude_start"] + self.consdb_visits["altitude_end"]) / 2
            else:
                altitude = self.consdb_visits["altitude_start"]
        else:
            altitude = pd.Series(np.nan, index=self.consdb_visits.index)

        return altitude

    @cached_property
    def airmass(self) -> pd.Series:
        if self._have_numeric_values("airmass"):
            airmass = super().airmass
        else:
            # Calculate airmass using Cassini's model.
            # Models the atmosphere with a uniform spherical shell with a
            # height 1/470 of the radius of the earth.
            # Described in Kirstensen https://doi.org/10.1002/asna.2123190313
            # and https://doi.org/10.2172/1574836
            a_cos_zd = 470 * np.cos(np.radians(self.zd))
            airmass = pd.Series(np.sqrt(a_cos_zd**2 + 941) - a_cos_zd, index=self.consdb_visits.index)

        return airmass

    @cached_property
    def shut_time(self) -> pd.Series:
        if self._have_numeric_values("shut_time"):
            shut_time = super().shut_time
        else:
            shut_time = pd.Series(SysEngVals().exptime, index=self.consdb_visits.index)

        return shut_time

    @cached_property
    def eff_time_m5(self):
        if self._have_numeric_values("eff_time_m5"):
            eff_time = super().eff_time_m5
        elif self._have_numeric_values("eff_time_median"):
            # From SMTN-002, and map to np.nan if not a standard band.
            ref_mags = defaultdict(
                np.array(np.nan).item,
                {"u": 23.70, "g": 24.97, "r": 24.52, "i": 24.13, "z": 23.56, "y": 22.55}.items(),
            )
            eff_time = self.band.map(ref_mags) + 1.25 * np.log10(self.consdb_visits["eff_time_median"] / 30)
        else:
            eff_time = pd.Series(np.nan, index=self.consdb_visits.index)

        return eff_time

    @cached_property
    def band(self) -> pd.Series:
        band = self.consdb_visits["band"].copy()
        missing_band: pd.Series[bool] = band.isna()
        band[missing_band] = self.consdb_visits.loc[missing_band, "physical_filter"].str.get(0)
        return band

    @cached_property
    def _sky_bg_column(self) -> str:
        return "sky_bg_median"

    @cached_property
    def _seeing_zenith_500nm_column(self) -> str:
        return "seeing_zenith_500nm_median"

    @cached_property
    def _psf_sigma_column(self) -> str:
        return "psf_sigma_median"

    @cached_property
    def atmospheric_extinction_with_clouds_and_air(self) -> pd.Series:
        """The extinction for each visit.

        Returns
        -------
        extinction : `pd.Series`
            The extinction of each visit, in magnitudes.
        """
        # If SysEngVals is correct, the difference between that at what is
        # measured will be the extinction.
        sys_band_zp_t = defaultdict(np.array(np.nan).item, SysEngVals().zp_t.items())
        sys_zp = self.band.map(sys_band_zp_t) + 2.5 * np.log10(self.shut_time)
        extinction = self.zero_point - sys_zp
        return extinction


class ComcamConsDBVisits(LSSTCamConsDBVisits):

    # Set "constant" properties using cached_property rather
    # than actual attributes, because abstract "members" can only be declared
    # as such using properties, and overriding a property with an actual
    # attribute confuses type checkers.

    @property
    def instrument(self) -> str:
        return "lsstcomcam"


class ComcamSimConsDBVisits(ComcamConsDBVisits):

    @property
    def instrument(self) -> str:
        return "lsstcomcamsim"


class AuxTelConsDBVisits(PachonConsDBVisits):

    @property
    def telescope(self) -> str:
        return "auxtel"


class LATISSConsDBVisits(AuxTelConsDBVisits):

    @property
    def instrument(self) -> str:
        return "latiss"

    @property
    def num_exposures(self) -> int:
        return 1

    @property
    def pixel_scale(self) -> pd.Series:
        """The pixel scale in asec/pixel.

        Returns
        -------
        pixel_scale : `pd.Series`
            The pixel scale.
        """
        if self._have_numeric_values("pixel_scale"):
            pixel_scale = super().pixel_scale
        else:
            # From TSTN-006
            pixel_scale = pd.Series(1.0, index=self.consdb_visits.index)
        return pixel_scale

    @cached_property
    def _sky_bg_column(self) -> str:
        return "sky_bg"

    @cached_property
    def _seeing_zenith_500nm_column(self) -> str:
        return "seeing_zenith_500nm"

    @cached_property
    def _psf_sigma_column(self) -> str:
        return "psf_sigma"

    @cached_property
    def _eff_time_column(self) -> str:
        return "eff_time"

    @cached_property
    def _zero_point_column(self) -> str:
        return "zero_point"

    @cached_property
    def eff_time_m5(self) -> pd.Series:
        """Effective exposure time, in seconds.
        This is the time to reach the m5 depth in each visit, under
        a reference set of conditions.

        Returns
        -------
        eff_time : `pd.Series`
            The effective exposure time.
        """
        if self._have_numeric_values(self._eff_time_column):
            eff_time = super().eff_time_m5
        else:
            eff_time = pd.Series(np.nan, index=self.consdb_visits.index)

        return eff_time


# Different subclasses of ConsDBOpsimConverter are needed for
# different instruments.
# Make a factory function to return the correct one.
def load_consdb_visits(instrument: str = "lsstcomcam", *args, **kwargs) -> ConsDBVisits:
    """Return visits from the consdb.

    Parameters
    ----------
    instrument : `str`, optional
        The name of the instrument, by default "lsstcomcamsim"

    Returns
    -------
    visits : `ConsDBVisits`
        Visits from the consdb.
    """

    if instrument not in KNOWN_INSTRUMENTS:
        # Yes, the "case: _" clause below would take care of this,
        # but having this explicitly here makes sure KNOWN_INSTRUMENTS
        # gets properly updated when new instruments are added.
        raise NotImplementedError

    match instrument:
        case "lsstcam":
            converter: ConsDBVisits = LSSTCamConsDBVisits(*args, **kwargs)
        case "lsstcomcam":
            converter: ConsDBVisits = ComcamConsDBVisits(*args, **kwargs)
        case "lsstcomcamsim":
            converter: ConsDBVisits = ComcamSimConsDBVisits(*args, **kwargs)
        case "latiss":
            converter: ConsDBVisits = LATISSConsDBVisits(*args, **kwargs)
        case _:
            raise NotImplementedError

    return converter



__all__ = (
    "BaseBasisFunction",
    "HealpixLimitedBasisFunctionMixin",
    "ConstantBasisFunction",
    "SimpleArrayBasisFunction",
    "DelayStartBasisFunction",
    "AvoidFastRevisitsBasisFunction",
    "VisitRepeatBasisFunction",
    "M5DiffBasisFunction",
    "M5DiffAtHpixBasisFunction",
    "StrictBandBasisFunction",
    "StrictFilterBasisFunction",
    "BandChangeBasisFunction",
    "FilterChangeBasisFunction",
    "SlewtimeBasisFunction",
    "CadenceEnhanceBasisFunction",
    "CadenceEnhanceTrapezoidBasisFunction",
    "AzimuthBasisFunction",
    "AzModuloBasisFunction",
    "DecModuloBasisFunction",
    "MapModuloBasisFunction",
    "SeasonCoverageBasisFunction",
    "NObsPerYearBasisFunction",
    "CadenceInSeasonBasisFunction",
    "NearSunHighAirmassBasisFunction",
    "EclipticBasisFunction",
    "VisitGap",
    "NGoodSeeingBasisFunction",
    "AvoidDirectWind",
    "BalanceVisits",
    "RewardNObsSequence",
    "BandDistBasisFunction",
    "FilterDistBasisFunction",
    "RewardRisingBasisFunction",
    "send_unused_deprecation_warning",
)

import warnings

import healpy as hp
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord

from rubin_scheduler.scheduler import features
from rubin_scheduler.scheduler.utils import IntRounded, get_current_footprint
from rubin_scheduler.skybrightness_pre import dark_m5
from rubin_scheduler.utils import DEFAULT_NSIDE, SURVEY_START_MJD, _hpid2_ra_dec


def send_unused_deprecation_warning(name):
    message = (
        f"The basis function {name} is not in use by the current "
        "baseline scheduler and may be deprecated shortly. "
        "Please contact the rubin_scheduler maintainers if "
        "this is in use elsewhere."
    )
    warnings.warn(message, FutureWarning)


class BaseBasisFunction:
    """Class that takes features and computes a reward function when
    called."""

    def __init__(self, nside=DEFAULT_NSIDE, bandname=None, filtername=None, **kwargs):
        # Set if basis function needs to be recalculated if there is a new
        # observation
        self.update_on_newobs = True
        # Set if basis function needs to be recalculated if conditions
        # change
        self.update_on_mjd = True
        # Dict to hold all the features we want to track
        self.survey_features = {}
        # Keep track of the last time the basis function was called.
        # If mjd doesn't change, use cached value
        self.mjd_last = None
        self.value = 0
        # list the attributes to compare to check if basis functions
        # are equal.
        self.attrs_to_compare = []
        # Do we need to recalculate the basis function
        self.recalc = True
        # Basis functions don't technically all need an nside, but so
        # many do might as well set it here
        self.nside = nside

        if filtername is not None:
            warnings.warn(
                "Use of `filtername` will be deprecated in favor of `bandname` at v4", FutureWarning
            )
            bandname = filtername
            # Save filtername as a backup in case someone tries to access it
            self.filtername = filtername
        self.bandname = bandname

    def add_observations_array(self, observations_array, observations_hpid):
        """Similar to add_observation, but for loading a whole
        array of observations at a time.

        Parameters
        ----------
        observations_array_in : `np.array`
            An array of completed observations (with columns like
            rubin_scheduler.scheduler.utils.ObservationArray).
            Should be sorted by MJD.
        observations_hpid_in : `np.array`
            Same as observations_array_in, but larger and with an
            additional column for HEALpix id.
            Each observation is listed multiple times,
            once for every HEALpix it overlaps.
        """

        for feature in self.survey_features:
            self.survey_features[feature].add_observations_array(observations_array, observations_hpid)
        if self.update_on_newobs:
            self.recalc = True

    def add_observation(self, observation, indx=None):
        """
        Parameters
        ----------
        observation : `np.array`
            An array with information about the input observation
        indx : `np.array`
            The indices of the healpix map that the observation overlaps
            with
        """
        for feature in self.survey_features:
            self.survey_features[feature].add_observation(observation, indx=indx)
        if self.update_on_newobs:
            self.recalc = True

    def check_feasibility(self, conditions):
        """If there is logic to decide if something is feasible
        (e.g., only if  moon is down), it can be calculated here.

        Helps prevent full __call__ from being called more than needed.
        """
        return True

    def _calc_value(self, conditions, **kwargs):
        self.value = 0
        # Update the last time we had an mjd
        self.mjd_last = conditions.mjd + 0
        self.recalc = False
        return self.value

    def __eq__(self):
        # XXX--to work on if we need to make a registry of basis functions.
        pass

    def __ne__(self):
        pass

    def __call__(self, conditions, **kwargs):
        """
        Parameters
        ----------
        conditions : `rubin_scheduler.scheduler.features.conditions` object
             Object that has attributes for all the current conditions.

        Return a reward healpix map or a reward scalar.
        """
        # If we are not feasible, return -inf
        if not self.check_feasibility(conditions):
            return -np.inf
        if self.recalc:
            self.value = self._calc_value(conditions, **kwargs)
        if self.update_on_mjd:
            if conditions.mjd != self.mjd_last:
                self.value = self._calc_value(conditions, **kwargs)
        return self.value

    def label(self):
        """Create a label for this basis function.

        Returns
        -------
        label : `str`
            A string suitable for labeling the basis function in a
            plot or table.
        """
        label = self.__class__.__name__.replace("BasisFunction", "")

        if self.bandname is not None:
            label += f" {self.bandname}"

        label += f" @{id(self)}"

        return label


class HealpixLimitedBasisFunctionMixin:
    """A mixin to limit a basis function to a set of Healpix pixels."""

    def __init__(self, hpid, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hpid = hpid

    def check_feasibility(self, conditions):
        """Check the feasibility of the current set of conditions.

        Parameters
        ----------
        conditions : `rubin_scheduler.scheduler.features.Conditions`
            The conditions for which to test feasibility.

        Returns
        -------
        feasibility : `bool`
            True if the current set of conditions is feasible, False
            otherwise.
        """

        if super().check_feasibility(conditions):
            if self.recalc or (self.update_on_mjd and conditions.mjd != self.mjd_last):
                value = self._calc_value(conditions)
            else:
                value = super().value

            feasibility = np.nanmax(value) > -np.inf
        else:
            feasibility = False
        return feasibility

    def _calc_value(self, conditions, all_sky=False, **kwargs):
        all_sky_value = super()._calc_value(conditions, **kwargs)

        if all_sky:
            return all_sky_value

        if np.isscalar(all_sky_value):
            value = all_sky_value
        else:
            assert len(all_sky_value) == hp.nside2npix(self.nside)
            value = all_sky_value[self.hpid]

        return value


class ConstantBasisFunction(BaseBasisFunction):
    """Just add a constant"""

    def __call__(self, conditions, **kwargs):
        return 1


class SimpleArrayBasisFunction(BaseBasisFunction):
    def __init__(self, value, *args, **kwargs):
        self.assigned_value = value
        super().__init__(*args, **kwargs)

    def _calc_value(self, conditions, **kwargs):
        self.value = self.assigned_value
        return self.value


class DelayStartBasisFunction(BaseBasisFunction):
    """Force things to not run before a given night.

    Parameters
    ----------
    nights_delay : `float`, optional
        Return False until conditions.night >= nights_delay.
    """

    def __init__(self, nights_delay=365.25 * 5):
        super().__init__()
        self.nights_delay = nights_delay

    def check_feasibility(self, conditions):
        result = True
        if conditions.night < self.nights_delay:
            result = False
        return result


class BandDistBasisFunction(BaseBasisFunction):
    """Track band distribution, increase reward as fraction of observations
    in specified band drops.
    """

    def __init__(self, bandname="r"):
        super(BandDistBasisFunction, self).__init__(bandname=bandname)

        self.survey_features = {}
        # Count of all the observations
        self.survey_features["n_obs_count_all"] = features.NObsCount(bandname=None)
        # Count in band
        self.survey_features["n_obs_count_in_filt"] = features.NObsCount(bandname=bandname)

    def _calc_value(self, conditions, indx=None):
        result = self.survey_features["n_obs_count_all"].feature / (
            self.survey_features["n_obs_count_in_filt"].feature + 1
        )
        return result


class FilterDistBasisFunction(BandDistBasisFunction):
    """Deprecated version of BandDistBasisFunction"""

    def __init__(self, filtername="r"):
        warnings.warn("FilterDistBasisFunction deprecated for BandDistBasisFunction", FutureWarning)
        super().__init__(bandname=filtername)


class NObsPerYearBasisFunction(BaseBasisFunction):
    """Reward areas that have not been observed N-times in the last year

    Parameters
    ----------
    bandname : `str` ('r')
        The band to track
    footprint : `np.array`
        Should be a HEALpix map. Values of 0 or np.nan will be ignored.
    n_obs : `int` (3)
        The number of observations to demand
    season : `float` (300)
        The amount of time to allow pass before marking a region as "behind".
        Default 365.25 (days).
    season_start_hour : `float` (-2)
        When to start the season relative to RA 180 degrees away
        from the sun (hours)
    season_end_hour : `float` (2)
        When to consider a season ending, the RA relative to the sun + 180
        degrees. (hours)
    night_max : float (365)
        Set value to zero after night_max is reached (days)
    """

    def __init__(
        self,
        bandname="r",
        nside=DEFAULT_NSIDE,
        footprint=None,
        n_obs=3,
        season=300,
        season_start_hour=-4.0,
        season_end_hour=2.0,
        night_max=365,
        filtername=None,
    ):
        if filtername is not None:
            warnings.warn("filtername deprecated in favor of bandname", FutureWarning)
            bandname = filtername
        super(NObsPerYearBasisFunction, self).__init__(nside=nside, bandname=bandname)
        self.footprint = footprint
        self.n_obs = n_obs
        self.season = season
        self.season_start_hour = (season_start_hour) * np.pi / 12.0  # To radians
        self.season_end_hour = season_end_hour * np.pi / 12.0  # To radians

        self.survey_features["last_n_mjds"] = features.LastNObsTimes(
            nside=nside, bandname=bandname, n_obs=n_obs
        )
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)
        self.out_footprint = np.where((footprint == 0) | np.isnan(footprint))
        self.night_max = night_max

    def _calc_value(self, conditions, indx=None):
        if conditions.night > self.night_max:
            return 0

        result = self.result.copy()
        behind_pix = np.where((conditions.mjd - self.survey_features["last_n_mjds"].feature[0]) > self.season)
        result[behind_pix] = 1

        # let's ramp up the weight depending on how far into the
        # observing season the healpix is
        mid_season_ra = (conditions.sun_ra + np.pi) % (2.0 * np.pi)
        # relative RA
        relative_ra = (conditions.ra - mid_season_ra) % (2.0 * np.pi)
        relative_ra = (self.season_end_hour - relative_ra) % (2.0 * np.pi)
        # ok, now
        relative_ra[
            np.where(IntRounded(relative_ra) > IntRounded(self.season_end_hour - self.season_start_hour))
        ] = 0

        weight = relative_ra / (self.season_end_hour - self.season_start_hour)
        result *= weight

        # mask off anything outside the footprint
        result[self.out_footprint] = 0

        return result


class NGoodSeeingBasisFunction(BaseBasisFunction):
    """Try to get N "good seeing" images each observing season.

    Parameters
    ----------
    bandname : `str`
        Bandpass in which to count images. Default r.
    nside : `int`
        The nside of the map for the basis function. Should match
        survey and scheduler nside.
        Default None uses `set_default_nside`.
    seeing_fwhm_max : `float`
        Value to consider as "good" threshold (arcsec).
        Default of 0.8 arcseconds.
    m5_penalty_max : `float`
        The maximum depth loss that is considered acceptable (magnitudes),
        compared to the dark-sky map in this band.
        Default 0.5 magnitudes.
    n_obs_desired : `int`
        Number of good seeing observations to collect per season.
        Default 3.
    mjd_start : `float`
        The starting MJD of the survey.
        Default None uses `rubin_scheduler.utils.SURVEY_START_MJD`.
    footprint : `np.array`, (N,)
        Only use area where footprint > 0. Should be a HEALpix map.
        Default None calls `get_current_footprint()`.
    """

    def __init__(
        self,
        bandname="r",
        nside=DEFAULT_NSIDE,
        seeing_fwhm_max=0.8,
        m5_penalty_max=0.5,
        n_obs_desired=3,
        mjd_start=None,
        footprint=None,
        filtername=None,
    ):
        if filtername is not None:
            warnings.warn("filtername deprecated in favor of bandname", FutureWarning)
            bandname = filtername
        super().__init__(nside=nside, bandname=bandname)
        self.seeing_fwhm_max = seeing_fwhm_max
        self.m5_penalty_max = m5_penalty_max
        self.n_obs_desired = n_obs_desired
        if mjd_start is None:
            mjd_start = SURVEY_START_MJD
        self.mjd_start = mjd_start
        self.survey_features["N_good_seeing"] = features.NObservationsCurrentSeason(
            bandname=self.bandname,
            mjd_start=self.mjd_start,
            seeing_fwhm_max=self.seeing_fwhm_max,
            m5_penalty_max=self.m5_penalty_max,
            nside=self.nside,
        )
        # Set footprint to current survey footprint class if undefined.
        if footprint is None:
            footprints, labels = get_current_footprint(self.nside)
            footprint = footprints[self.bandname]
        self.footprint = footprint
        self.result = np.zeros(hp.nside2npix(self.nside))
        self.dark_map = None

    def _calc_value(self, conditions, indx=None):
        if self.bandname is not None:
            if self.dark_map is None:
                self.dark_map = dark_m5(
                    conditions.dec, self.bandname, conditions.site.latitude_rad, fiducial_FWHMEff=0.7
                )
        # Return the same kind of array (not float) regardless
        # of result
        result = self.result.copy()
        # Update the feature to the current time.
        self.survey_features["N_good_seeing"].season_update(conditions=conditions)

        m5_penalty = self.dark_map - conditions.m5_depth[self.bandname]
        potential_pixels = np.where(
            (m5_penalty <= self.m5_penalty_max)
            & (conditions.fwhm_eff[self.bandname] <= self.seeing_fwhm_max)
            & (self.survey_features["N_good_seeing"].feature < self.n_obs_desired)
            & (self.footprint > 0)
        )[0]

        result[potential_pixels] = 1
        return result


def az_rel_point(azs, point_az):
    az_rel_moon = (azs - point_az) % (2.0 * np.pi)
    if isinstance(azs, np.ndarray):
        over = np.where(az_rel_moon > np.pi)
        az_rel_moon[over] = 2.0 * np.pi - az_rel_moon[over]
    else:
        if az_rel_moon > np.pi:
            az_rel_moon = 2.0 * np.pi - az_rel_moon
    return az_rel_moon


class EclipticBasisFunction(BaseBasisFunction):
    """Mark the area around the ecliptic"""

    def __init__(self, nside=DEFAULT_NSIDE, distance_to_eclip=25.0):
        super(EclipticBasisFunction, self).__init__(nside=nside)
        self.distance_to_eclip = np.radians(distance_to_eclip)
        ra, dec = _hpid2_ra_dec(nside, np.arange(hp.nside2npix(self.nside)))
        self.result = np.zeros(ra.size)
        coord = SkyCoord(ra=ra * u.rad, dec=dec * u.rad)
        eclip_lat = coord.barycentrictrueecliptic.lat.radian
        good = np.where(np.abs(eclip_lat) < self.distance_to_eclip)
        self.result[good] += 1

    def __call__(self, conditions, indx=None):
        return self.result


class CadenceInSeasonBasisFunction(BaseBasisFunction):
    """Drive observations at least every N days in a given area

    Parameters
    ----------
    drive_map : `np.ndarray`, (N,)
        A HEALpix map with values of 1 where the cadence should be driven.
    bandname : `str`
        The bands that can count.
    season_span : `float`
        How long to consider a spot "in_season" (hours).
    cadence : `float`
        How long to wait before activating the basis function (days).
    """

    def __init__(
        self, drive_map, bandname="griz", season_span=2.5, cadence=2.5, nside=DEFAULT_NSIDE, filtername=None
    ):
        if filtername is not None:
            warnings.warn("filtername deprecated in favor of bandname", FutureWarning)
            bandname = filtername
        super(CadenceInSeasonBasisFunction, self).__init__(nside=nside, bandname=bandname)
        self.drive_map = drive_map
        self.season_span = season_span / 12.0 * np.pi  # To radians
        self.cadence = cadence
        self.survey_features["last_observed"] = features.LastObserved(nside=nside, bandname=bandname)
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()
        ra_mid_season = (conditions.sunRA + np.pi) % (2.0 * np.pi)

        angle_to_mid_season = np.abs(conditions.ra - ra_mid_season)
        over = np.where(IntRounded(angle_to_mid_season) > IntRounded(np.pi))
        angle_to_mid_season[over] = 2.0 * np.pi - angle_to_mid_season[over]

        days_lag = conditions.mjd - self.survey_features["last_observed"].feature

        active_pix = np.where(
            (IntRounded(days_lag) >= IntRounded(self.cadence))
            & (self.drive_map == 1)
            & (IntRounded(angle_to_mid_season) < IntRounded(self.season_span))
        )
        result[active_pix] = 1.0

        return result


class SeasonCoverageBasisFunction(BaseBasisFunction):
    """Basis function to encourage N observations per observing season.

    Parameters
    ----------
    bandname : `str`, optional
        Count observations in this band. Default 'r'.
    nside : `int`, optional
        Nside for the healpix map to use for the feature.
        This should match the nside of the survey and scheduler.
    footprint : `np.array` (N,), optional
        Healpix map of the footprint where one should demand coverage
        every season. Default None will call `get_current_footprint()`.
    n_per_season : `int`, optional
        The number of observations to attempt to gather every season.
        Default of 3 is suitable for first year template building.
    mjd_start : `float`, optional
        The mjd of the start of the survey (days).
        Default None uses `rubin_scheduler.utils.SURVEY_START_MJD`.
    season_frac_start : `float`
        Only start trying to gather observations after a season
        is fractionally this far along.
        Seasons start when the apparent position of sun is at the RA of
        the pixel (0) and finish when the sun returns again to this RA.
        The default of 0.5 means that the basis function will not
        start returning values until the RA reaches the peak of its season.
    """

    def __init__(
        self,
        bandname="r",
        nside=DEFAULT_NSIDE,
        footprint=None,
        n_per_season=3,
        mjd_start=None,
        season_frac_start=0.5,
        filtername=None,
    ):
        if filtername is not None:
            warnings.warn("filtername deprecated in favor of bandname", FutureWarning)
            bandname = filtername
        send_unused_deprecation_warning("SeasonCoverageBasisFunction")
        super().__init__(nside=nside, bandname=bandname)

        if footprint is None:
            footprints, labels = get_current_footprint(self.nside)
            footprint = footprints[self.bandname]
        self.footprint = footprint
        # Calculate the RA values for each spot on the footprint
        ra, dec = _hpid2_ra_dec(nside, np.arange(hp.nside2npix(nside)))
        self.ra_deg = np.degrees(ra)

        self.n_per_season = n_per_season
        if mjd_start is None:
            mjd_start = SURVEY_START_MJD
        self.mjd_start = mjd_start
        self.season_frac_start = season_frac_start
        # Track how many observations have been taken at each RA/Dec
        # in the current observing season (for that point on the sky).
        self.survey_features["n_obs_season"] = features.NObservationsCurrentSeason(
            bandname=bandname, nside=nside, mjd_start=mjd_start
        )
        self.result = np.zeros(hp.nside2npix(self.nside), dtype=float)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()

        # Update the feature to the current time
        self.survey_features["n_obs_season"].season_update(conditions=conditions)
        feature = self.survey_features["n_obs_season"].feature

        # Get the season from the conditions object.
        season = conditions.season

        # Evaluate where there are not yet enough observations and also
        # that season is far enough along to require more weight.
        not_enough = np.where(
            (self.footprint > 0)
            & (feature < self.n_per_season)
            & ((IntRounded(season - np.floor(season)) > IntRounded(self.season_frac_start)))
        )
        result[not_enough] = 1
        return result


class AvoidFastRevisitsBasisFunction(BaseBasisFunction):
    """Marks targets as unseen if they are in a specified time window
    in order to avoid fast revisits.

    Parameters
    ----------
    bandname : `str` or None
        The name of the band for this target map.
        Using None will match visits in any band.
    gap_min : `float`
        Minimum time for the gap (minutes).
    nside: `int` or None
        The healpix resolution.
    penalty_val : `float`
        The reward value to use for regions to penalize.
        Will be masked if set to np.nan (default).
    """

    def __init__(self, bandname="r", nside=DEFAULT_NSIDE, gap_min=25.0, penalty_val=np.nan, filtername=None):
        if filtername is not None:
            warnings.warn("filtername deprecated in favor of bandname", FutureWarning)
            bandname = filtername
        super().__init__(nside=nside, bandname=bandname)

        self.bandname = bandname
        self.penalty_val = penalty_val

        self.gap_min = IntRounded(gap_min / 60.0 / 24.0)
        self.nside = nside

        self.survey_features = dict()
        self.survey_features["Last_observed"] = features.LastObserved(bandname=bandname, nside=nside, fill=0)

    def _calc_value(self, conditions, indx=None):
        result = np.ones(hp.nside2npix(self.nside), dtype=float)
        diff = IntRounded(conditions.mjd - self.survey_features["Last_observed"].feature)
        bad = np.where(diff < self.gap_min)[0]
        result[bad] = self.penalty_val
        return result


class NearSunHighAirmassBasisFunction(BaseBasisFunction):
    """Reward areas on the sky at high airmass, within 90 degrees azimuth
    of the Sun, such as suitable for the near-sun twilight microsurvey for
    near- or interior-to earth asteroids.

    Parameters
    ----------
    nside : `int`, optional
        Nside for the basis function. If None, uses `set_default_nside()`.
    max_airmass : `float`, oprionl
        The maximum airmass to try and observe (unitless).
    penalty : `float`, optional
        The value to fill in non-rewarded parts of the sky.
        Default np.nan, which serves to mask regions exceeding the airmass
        limit and more than 90 degrees azimuth toward the sun.
    """

    def __init__(self, nside=DEFAULT_NSIDE, max_airmass=2.5, penalty=np.nan):
        super().__init__(nside=nside)
        self.max_airmass = IntRounded(max_airmass)
        self.result = np.empty(hp.nside2npix(self.nside))
        self.result.fill(penalty)

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()
        valid_airmass = np.where(np.isfinite(conditions.airmass))[0]
        good_pix = np.where(
            (conditions.airmass[valid_airmass] >= 1.0)
            & (IntRounded(conditions.airmass[valid_airmass]) < self.max_airmass)
            & (IntRounded(np.abs(conditions.az_to_sun[valid_airmass])) < IntRounded(np.pi / 2.0))
        )
        result[valid_airmass[good_pix]] = (
            conditions.airmass[valid_airmass][good_pix] / self.max_airmass.initial
        )
        return result


class VisitRepeatBasisFunction(BaseBasisFunction):
    """
    Basis function to reward re-visiting an area on the sky.
    Looking for Solar System objects.

    Parameters
    ----------
    gap_min : `float` (15.)
        Minimum time for the gap (minutes)
    gap_max : `float` (45.)
        Maximum time for a gap
    bandname : `str` ('r')
        The band(s) to count with pairs
    npairs : `int` (1)
        The number of pairs of observations to attempt to gather
    """

    def __init__(
        self, gap_min=25.0, gap_max=45.0, bandname="r", nside=DEFAULT_NSIDE, npairs=1, filtername=None
    ):
        if filtername is not None:
            warnings.warn("filtername deprecated in favor of bandname", FutureWarning)
            bandname = filtername
        super(VisitRepeatBasisFunction, self).__init__(nside=nside, bandname=bandname)

        self.gap_min = IntRounded(gap_min / 60.0 / 24.0)
        self.gap_max = IntRounded(gap_max / 60.0 / 24.0)
        self.npairs = npairs
        self.survey_features = {}
        # Track the number of pairs that have been taken in a night
        self.survey_features["Pair_in_night"] = features.PairInNight(
            bandname=bandname, gap_min=gap_min, gap_max=gap_max, nside=nside
        )

        # When was it last observed
        # XXX--since this feature is also in Pair_in_night, I should just
        # access that one!
        self.survey_features["Last_observed"] = features.LastObserved(bandname=bandname, nside=nside)

    def _calc_value(self, conditions, indx=None):
        result = np.zeros(hp.nside2npix(self.nside), dtype=float)
        if indx is None:
            indx = np.arange(result.size)
        diff = conditions.mjd - self.survey_features["Last_observed"].feature[indx]
        mask = np.isnan(diff)
        # remove NaNs from diff, but save mask so we exclude those values
        # later.
        diff[mask] = 0.0
        ir_diff = IntRounded(diff)
        good = np.where(
            (ir_diff >= self.gap_min)
            & (ir_diff <= self.gap_max)
            & (self.survey_features["Pair_in_night"].feature[indx] < self.npairs)
            & (~mask)
        )[0]
        result[indx[good]] += 1.0
        return result


class M5DiffBasisFunction(BaseBasisFunction):
    """Basis function based on the 5-sigma depth.
    Look up the best depth a healpixel achieves, and compute
    the limiting depth difference given current conditions

    Parameters
    ----------
    bandname : `str`, optional
        The band to consider for visits.
    fiducial_FWHMEff : `float`, optional
        The zenith seeing to assume for "good" conditions.
        While the dark sky depth map simply scales with this value,
        picking a reasonable fiducial_FWHMEff is important because
        this effects the overall value and scale of the reward
        from the basis function.
    nside : `int`, optional
        The nside for the basis function.
        Default None uses `set_default_nside()`.
    apply_cloud_extinction : `bool`
        Apply extinction from cloud maps. Default False.
    """

    def __init__(
        self,
        bandname="r",
        fiducial_FWHMEff=0.7,
        nside=DEFAULT_NSIDE,
        filtername=None,
        apply_cloud_extinction=False,
    ):
        if filtername is not None:
            warnings.warn("filtername deprecated in favor of bandname", FutureWarning)
            bandname = filtername
        super().__init__(nside=nside, bandname=bandname)
        # The dark sky surface brightness values
        self.dark_map = None
        self.fiducial_FWHMEff = fiducial_FWHMEff
        self.bandname = bandname
        self.apply_cloud_extinction = apply_cloud_extinction

    def _calc_value(self, conditions, indx=None):
        if self.dark_map is None:
            self.dark_map = dark_m5(
                conditions.dec, self.bandname, conditions.site.latitude_rad, self.fiducial_FWHMEff
            )

        result = conditions.m5_depth[self.bandname] - self.dark_map
        if self.apply_cloud_extinction:
            if conditions.cloud_maps is not None:
                extinction = conditions.cloud_maps.extinction_closest(conditions.mjd)
                result -= extinction
        return result


class M5DiffAtHpixBasisFunction(HealpixLimitedBasisFunctionMixin, M5DiffBasisFunction):
    pass


class StrictBandBasisFunction(BaseBasisFunction):
    """Remove the bonus for staying in the same band
    if certain conditions are met.

    If the moon rises/sets or twilight starts/ends, it makes a lot of sense
    to consider a band change. This basis function rewards if it matches
    the current band, the moon rises or sets, twilight starts or stops,
    or there has been a large gap since the last observation.

    Parameters
    ----------
    time_lag : `float` (10.)
        If there is a gap between observations longer than this,
        let the band change (minutes)
    twi_change : `float` (-18.)
        The sun altitude to consider twilight starting/ending (degrees)
    note_free : `str` ('DD')
        No penalty for changing bands if the last observation note field
        includes `note_free` string.
        Useful for giving a free band change after deep drilling sequence
    """

    def __init__(self, time_lag=10.0, bandname="r", twi_change=-18.0, note_free="DD"):
        super(StrictBandBasisFunction, self).__init__(bandname=bandname)

        self.time_lag = time_lag / 60.0 / 24.0  # Convert to days
        self.twi_change = np.radians(twi_change)

        self.survey_features = {}
        self.survey_features["Last_observation"] = features.LastObservation()
        self.note_free = note_free

    def _calc_value(self, conditions, **kwargs):
        # Did the moon set or rise since last observation?
        moon_changed = conditions.moon_alt * self.survey_features["Last_observation"].feature["moonAlt"] < 0

        # Are we already in the band (or at start of night)?
        in_band = (conditions.current_band == self.bandname) | (conditions.current_band is None)

        # Has enough time past?
        time_past = IntRounded(
            conditions.mjd - self.survey_features["Last_observation"].feature["mjd"]
        ) > IntRounded(self.time_lag)

        # Did twilight start/end?
        twi_changed = (conditions.sun_alt - self.twi_change) * (
            self.survey_features["Last_observation"].feature["sunAlt"] - self.twi_change
        ) < 0

        # Did we just finish a DD sequence
        was_dd = self.note_free in self.survey_features["Last_observation"].feature["scheduler_note"]

        # Is the band mounted?
        mounted = self.bandname in conditions.mounted_bands

        if (moon_changed | in_band | time_past | twi_changed | was_dd) & mounted:
            result = 1.0
        else:
            result = 0.0

        return result


class StrictFilterBasisFunction(StrictBandBasisFunction):
    """Deprecated in favor of StrictBandBasisFunction"""

    def __init__(self, time_lag=10.0, filtername="r", twi_change=-18.0, note_free="DD"):
        warnings.warn(
            "StrictFilterBasisFunction deprecated in favor of StrictBandBasisFunction", FutureWarning
        )
        super().__init__(time_lag=time_lag, bandname=filtername, twi_change=twi_change, note_free=note_free)


class BandChangeBasisFunction(BaseBasisFunction):
    """Reward staying in the current band."""

    def __init__(self, bandname="r"):
        super(BandChangeBasisFunction, self).__init__(bandname=bandname)

    def _calc_value(self, conditions, **kwargs):
        if (conditions.current_band == self.bandname) | (conditions.current_band is None):
            result = 1.0
        else:
            result = 0.0
        return result


class FilterChangeBasisFunction(BandChangeBasisFunction):
    """Deprecated in favor of BandChangeBasisFunction"""

    def __init__(self, filtername="r"):
        warnings.warn(
            "FilterChangeBasisFunction deprecated in favor of BandChangeBasisFunction", FutureWarning
        )
        super().__init__(bandname=filtername)


class SlewtimeBasisFunction(BaseBasisFunction):
    """Reward slews that take little time

    Parameters
    ----------
    max_time : `float`
         The estimated maximum slewtime (seconds).
         Used to normalize so the basis function spans ~ -1-0
         in reward units. Default 135 seconds corresponds to just
         slightly less than a band change.
    bandname : `str` or None, optional
        The band to check for pre-post slewtime estimates.
        If the band is None, then bandpasses changes are NOT considered
        when calculating slewtime.
        If a slew includes a band change, other basis functions will
        decide on the reward, so the result here can be 0.
    nside : `int`, optional
        Nside for the basis function.
        Default None will use `set_default_nside()`.
    """

    def __init__(self, max_time=135.0, bandname="r", nside=DEFAULT_NSIDE, filtername=None):
        if filtername is not None:
            warnings.warn("filtername deprecated in favor of bandname", FutureWarning)
            bandname = filtername
        super(SlewtimeBasisFunction, self).__init__(nside=nside, bandname=bandname)

        self.maxtime = max_time
        self.nside = nside
        self.bandname = bandname

    def _calc_value(self, conditions, indx=None):
        # If we are in a different band, the
        # BandChangeBasisFunction will take it
        # But we can still use the MASK returned by
        # the slewtime map to remove inaccessible parts of the sky
        if conditions.current_band != self.bandname and self.bandname is not None:
            if np.size(conditions.slewtime) > 1:
                result = np.where(np.isfinite(conditions.slewtime), 0, np.nan)
            else:
                result = 0
        else:
            # Need to make sure smaller slewtime is larger reward.
            if np.size(conditions.slewtime) > 1:
                # Slewtime map can contain nans and/or
                # infs - mask these with nans
                result = np.where(
                    np.isfinite(conditions.slewtime),
                    -conditions.slewtime / self.maxtime,
                    np.nan,
                )
            else:
                result = -conditions.slewtime / self.maxtime
        return result


class CadenceEnhanceBasisFunction(BaseBasisFunction):
    """Drive a certain cadence

    Parameters
    ----------
    bandname : `str` ('gri')
        The band(s) that should be grouped together
    supress_window : `list` of `float`
        The start and stop window for when observations should be repressed
        (days)
    apply_area : healpix map
        The area over which to try and drive the cadence.
        Good values as 1, no cadence drive 0.
        Probably works as a bool array too.
    """

    def __init__(
        self,
        bandname="gri",
        nside=DEFAULT_NSIDE,
        supress_window=[0, 1.8],
        supress_val=-0.5,
        enhance_window=[2.1, 3.2],
        enhance_val=1.0,
        apply_area=None,
        filtername=None,
    ):
        if filtername is not None:
            warnings.warn("filtername deprecated in favor of bandname", FutureWarning)
            bandname = filtername
        super(CadenceEnhanceBasisFunction, self).__init__(nside=nside, bandname=bandname)

        self.supress_window = np.sort(supress_window)
        self.supress_val = supress_val
        self.enhance_window = np.sort(enhance_window)
        self.enhance_val = enhance_val

        self.survey_features = {}
        self.survey_features["last_observed"] = features.LastObserved(bandname=bandname)

        self.empty = np.zeros(hp.nside2npix(self.nside), dtype=float)
        # No map, try to drive the whole area
        if apply_area is None:
            self.apply_indx = np.arange(self.empty.size)
        else:
            self.apply_indx = np.where(apply_area != 0)[0]

    def _calc_value(self, conditions, indx=None):
        # copy an empty array
        result = self.empty.copy()
        if indx is not None:
            ind = np.intersect1d(indx, self.apply_indx)
        else:
            ind = self.apply_indx
        if np.size(ind) == 0:
            result = 0
        else:
            mjd_diff = conditions.mjd - self.survey_features["last_observed"].feature[ind]
            to_supress = np.where(
                (IntRounded(mjd_diff) > IntRounded(self.supress_window[0]))
                & (IntRounded(mjd_diff) < IntRounded(self.supress_window[1]))
            )
            result[ind[to_supress]] = self.supress_val
            to_enhance = np.where(
                (IntRounded(mjd_diff) > IntRounded(self.enhance_window[0]))
                & (IntRounded(mjd_diff) < IntRounded(self.enhance_window[1]))
            )
            result[ind[to_enhance]] = self.enhance_val
        return result


# https://docs.astropy.org/en/stable/_modules/astropy/modeling
# functional_models.html#Trapezoid1D
def trapezoid(x, amplitude, x_0, width, slope):
    """One dimensional Trapezoid model function"""
    # Compute the four points where the trapezoid changes slope
    # x1 <= x2 <= x3 <= x4
    x2 = x_0 - width / 2.0
    x3 = x_0 + width / 2.0
    x1 = x2 - amplitude / slope
    x4 = x3 + amplitude / slope

    result = x * 0

    # Compute model values in pieces between the change points
    range_a = np.logical_and(x >= x1, x < x2)
    range_b = np.logical_and(x >= x2, x < x3)
    range_c = np.logical_and(x >= x3, x < x4)

    result[range_a] = slope * (x[range_a] - x1)
    result[range_b] = amplitude
    result[range_c] = slope * (x4 - x[range_c])

    return result


class CadenceEnhanceTrapezoidBasisFunction(BaseBasisFunction):
    """Drive a certain cadence, like CadenceEnhanceBasisFunction
    but with smooth transitions

    Parameters
    ----------
    bandname : `str` ('gri')
        The band(s) that should be grouped together

    XXX--fill out doc string!
    """

    def __init__(
        self,
        bandname="gri",
        nside=DEFAULT_NSIDE,
        delay_width=2,
        delay_slope=2.0,
        delay_peak=0,
        delay_amp=0.5,
        enhance_width=3.0,
        enhance_slope=2.0,
        enhance_peak=4.0,
        enhance_amp=1.0,
        apply_area=None,
        season_limit=None,
        filtername=None,
    ):
        if filtername is not None:
            warnings.warn("filtername deprecated in favor of bandname", FutureWarning)
            bandname = filtername
        super(CadenceEnhanceTrapezoidBasisFunction, self).__init__(nside=nside, bandname=bandname)

        self.delay_width = delay_width
        self.delay_slope = delay_slope
        self.delay_peak = delay_peak
        self.delay_amp = delay_amp
        self.enhance_width = enhance_width
        self.enhance_slope = enhance_slope
        self.enhance_peak = enhance_peak
        self.enhance_amp = enhance_amp

        self.season_limit = season_limit / 12 * np.pi  # To radians

        self.survey_features = {}
        self.survey_features["last_observed"] = features.LastObserved(bandname=bandname)

        self.empty = np.zeros(hp.nside2npix(self.nside), dtype=float)
        # No map, try to drive the whole area
        if apply_area is None:
            self.apply_indx = np.arange(self.empty.size)
        else:
            self.apply_indx = np.where(apply_area != 0)[0]

    def suppress_enhance(self, x):
        result = x * 0
        result -= trapezoid(x, self.delay_amp, self.delay_peak, self.delay_width, self.delay_slope)
        result += trapezoid(
            x,
            self.enhance_amp,
            self.enhance_peak,
            self.enhance_width,
            self.enhance_slope,
        )

        return result

    def season_len(self, conditions):
        ra_mid_season = (conditions.sunRA + np.pi) % (2.0 * np.pi)
        angle_to_mid_season = np.abs(conditions.ra - ra_mid_season)
        over = np.where(IntRounded(angle_to_mid_season) > IntRounded(np.pi))
        angle_to_mid_season[over] = 2.0 * np.pi - angle_to_mid_season[over]

        return angle_to_mid_season

    def _calc_value(self, conditions, indx=None):
        # copy an empty array
        result = self.empty.copy()
        if indx is not None:
            ind = np.intersect1d(indx, self.apply_indx)
        else:
            ind = self.apply_indx
        if np.size(ind) == 0:
            result = 0
        else:
            mjd_diff = conditions.mjd - self.survey_features["last_observed"].feature[ind]
            result[ind] += self.suppress_enhance(mjd_diff)

        if self.season_limit is not None:
            radians_to_midseason = self.season_len(conditions)
            outside_season = np.where(radians_to_midseason > self.season_limit)
            result[outside_season] = 0

        return result


class AzimuthBasisFunction(BaseBasisFunction):
    """Reward staying in the same azimuth range.
    Possibly better than using slewtime, especially when selecting a
    large area of sky.
    """

    def __init__(self, nside=DEFAULT_NSIDE):
        send_unused_deprecation_warning("AzimuthBasisFunction")
        super(AzimuthBasisFunction, self).__init__(nside=nside)

    def _calc_value(self, conditions, indx=None):
        az_dist = conditions.az - conditions.tel_az
        az_dist = az_dist % (2.0 * np.pi)
        over = np.where(az_dist > np.pi)
        az_dist[over] = 2.0 * np.pi - az_dist[over]
        # Normalize sp between 0 and 1
        result = az_dist / np.pi
        return result


class AzModuloBasisFunction(BaseBasisFunction):
    """Try to replicate the Rothchild et al cadence forcing by
    only observing on limited az ranges per night.

    Parameters
    ----------
    az_limits : `list` of `float` pairs (None)
        The azimuth limits (degrees) to use.
    """

    def __init__(self, nside=DEFAULT_NSIDE, az_limits=None, out_of_bounds_val=-1.0):
        super(AzModuloBasisFunction, self).__init__(nside=nside)
        self.result = np.ones(hp.nside2npix(self.nside))
        if az_limits is None:
            spread = 100.0 / 2.0
            self.az_limits = np.radians(
                [
                    [360 - spread, spread],
                    [90.0 - spread, 90.0 + spread],
                    [180.0 - spread, 180.0 + spread],
                ]
            )
        else:
            self.az_limits = np.radians(az_limits)
        self.mod_val = len(self.az_limits)
        self.out_of_bounds_val = out_of_bounds_val

    def _calc_value(self, conditions, indx=None):
        result = self.result.copy()
        az_lim = self.az_limits[np.max(conditions.night) % self.mod_val]

        if az_lim[0] < az_lim[1]:
            out_pix = np.where(
                (IntRounded(conditions.az) < IntRounded(az_lim[0]))
                | (IntRounded(conditions.az) > IntRounded(az_lim[1]))
            )
        else:
            out_pix = np.where(
                (IntRounded(conditions.az) < IntRounded(az_lim[0]))
                | (IntRounded(conditions.az) > IntRounded(az_lim[1]))
            )[0]
        result[out_pix] = self.out_of_bounds_val
        return result


class DecModuloBasisFunction(BaseBasisFunction):
    """Emphasize dec bands on a nightly varying basis

    Parameters
    ----------
    dec_limits : `list` of `float` pairs (None)
        The azimuth limits (degrees) to use.
    """

    def __init__(self, nside=DEFAULT_NSIDE, dec_limits=None, out_of_bounds_val=-1.0):
        super(DecModuloBasisFunction, self).__init__(nside=nside)

        npix = hp.nside2npix(nside)
        hpids = np.arange(npix)
        ra, dec = _hpid2_ra_dec(nside, hpids)

        self.results = []

        if dec_limits is None:
            self.dec_limits = np.radians([[-90.0, -32.8], [-32.8, -12.0], [-12.0, 35.0]])
        else:
            self.dec_limits = np.radians(dec_limits)
        self.mod_val = len(self.dec_limits)
        self.out_of_bounds_val = out_of_bounds_val

        for limits in self.dec_limits:
            good = np.where((dec >= limits[0]) & (dec < limits[1]))[0]
            tmp = np.zeros(npix)
            tmp[good] = 1
            self.results.append(tmp)

    def _calc_value(self, conditions, indx=None):
        night_index = np.max(conditions.night % self.mod_val)
        result = self.results[night_index]

        return result


class MapModuloBasisFunction(BaseBasisFunction):
    """Similar to Dec_modulo, but now use input masks

    Parameters
    ----------
    inmaps : `list` of hp arrays
    """

    def __init__(self, inmaps):
        nside = hp.npix2nside(np.size(inmaps[0]))
        super(MapModuloBasisFunction, self).__init__(nside=nside)
        self.maps = inmaps
        self.mod_val = len(inmaps)

    def _calc_value(self, conditions, indx=None):
        indx = np.max(conditions.night % self.mod_val)
        result = self.maps[indx]
        return result


class VisitGap(BaseBasisFunction):
    """Basis function to create a visit gap based on the survey note field.

    Parameters
    ----------
    note : `str`
        Value of the observation "scheduler_note" field to be masked.
    band_names : list [str], optional
        List of band names that will be considered when evaluating
        if the gap has passed.
    gap_min : float (optional)
        Time gap (default=25, in minutes).
    penalty_val : float or np.nan
        Value of the penalty to apply (default is np.nan).

    Notes
    -----
    When a list of bands is provided, all bands must be observed before
    the gap requirement will be activated, and once activated, only
    observations in these bands will be evaluated in context of whether
    the last observation was at least gap in the past.
    """

    def __init__(self, note, band_names=None, gap_min=25.0, penalty_val=np.nan, filter_names=None):
        if filter_names is not None:
            warnings.warn("filter_names deprecated in favor of band_names", FutureWarning)
            band_names = filter_names
        super().__init__()
        self.penalty_val = penalty_val

        self.gap = gap_min / 60.0 / 24.0
        self.band_names = band_names

        self.survey_features = dict()
        if self.band_names is not None:
            for bandname in self.band_names:
                self.survey_features[f"LastObservationMjd::{bandname}"] = features.LastObservationMjd(
                    note=note, bandname=bandname
                )
        else:
            self.survey_features["LastObservationMjd"] = features.LastObservationMjd(scheduler_note=note)

    def check_feasibility(self, conditions):
        notes_last_observed = [last_observed.feature for last_observed in self.survey_features.values()]

        if any([last_observed is None for last_observed in notes_last_observed]):
            return True

        after_gap = [conditions.mjd - last_observed > self.gap for last_observed in notes_last_observed]

        return all(after_gap)

    def _calc_value(self, conditions, indx=None):
        return 1.0 if self.check_feasibility(conditions) else self.penalty_val


class AvoidDirectWind(BaseBasisFunction):
    """Mask the sky where the wind pressure exceeds `wind_speed_maximum`.

    Parameters
    ----------
    wind_speed_maximum : `float`, optional
        Wind speed to mark regions as unobservable (in m/s).
    nside : `int`, optional
        The nside for the basis function. Default None uses
        `set_default_nside()`.
    """

    def __init__(self, wind_speed_maximum=20.0, nside=DEFAULT_NSIDE):
        super().__init__(nside=nside)

        self.wind_speed_maximum = wind_speed_maximum

    def _calc_value(self, conditions, indx=None):
        reward_map = np.zeros(hp.nside2npix(self.nside))

        if conditions.wind_speed is None or conditions.wind_direction is None:
            return reward_map

        wind_pressure = conditions.wind_speed * np.cos(conditions.az - conditions.wind_direction)

        reward_map -= wind_pressure**2.0

        mask = wind_pressure > self.wind_speed_maximum

        reward_map[mask] = np.nan

        return reward_map


class BalanceVisits(BaseBasisFunction):
    """Balance visits across multiple surveys.

    Parameters
    ----------
    nobs_reference : `int`
        Expected number of observations across all interested surveys.
    note_survey : `str`
        Note value for the current survey.
    note_interest : `str`
        Substring with the name of interested surveys to be accounted.
    nside : `int`
        Healpix map resolution.

    Notes
    -----
    This basis function is designed to balance the reward of a group of
    surveys, such that the group get a reward boost based on the required
    collective number of observations.

    For example, if you have 3 surveys (e.g. SURVEY_A_REGION_1,
    SURVEY_A_REGION_2, SURVEY_A_REGION_3), when one of them is observed
    once (SURVEY_A_REGION_1) they all get a small reward boost proportional
    to the collective number of observations (`nobs_reference`). Further
    observations of SURVEY_A_REGION_1 would now cause the other surveys
    to gain a reward boost in relative to it.
    """

    def __init__(self, nobs_reference, note_survey, note_interest, nside=DEFAULT_NSIDE):
        super().__init__(nside=nside)

        self.nobs_reference = nobs_reference

        self.survey_features = {}
        self.survey_features["n_obs_survey"] = features.NObsCount(scheduler_note=note_survey)
        self.survey_features["n_obs_survey_interest"] = features.NObsCount(scheduler_note=note_interest)

    def _calc_value(self, conditions, indx=None):
        return (1 + np.floor(self.survey_features["n_obs_survey_interest"].feature / self.nobs_reference)) / (
            self.survey_features["n_obs_survey"].feature
            if self.survey_features["n_obs_survey"].feature > 0
            else 1
        )


class RewardNObsSequence(BaseBasisFunction):
    """Reward taking a sequence of observations.

    Parameters
    ----------
    n_obs_survey : `int`
        Number of observations to reward.
    note_survey : `str`
        The value of the observation note, to take into account.
    nside : `int`, optional
        Healpix map resolution (ignored).

    Notes
    -----
    This basis function is useful when a survey is composed of more than
    one observation (e.g. in different bands) and one wants to make sure
    they are all taken together.
    If the sequence is programmed into the FieldSurvey, this isn't necessary.
    """

    def __init__(self, n_obs_survey, note_survey, nside=DEFAULT_NSIDE):
        super().__init__(nside=nside)

        self.n_obs_survey = n_obs_survey

        self.survey_features = {}
        self.survey_features["n_obs_survey"] = features.NObsCount(scheduler_note=note_survey)

    def _calc_value(self, conditions, indx=None):
        return self.survey_features["n_obs_survey"].feature % self.n_obs_survey


class RewardRisingBasisFunction(BaseBasisFunction):
    """Reward parts of the sky that are rising.
    Optionally, mask out parts of the sky that are not rising.

    This produces a reward that increases
    as the field rises toward zenith, then abruptly
    falls as the field passes zenith.
    Negative hour angles (or hour angles > 180 degrees)
    indicate a rising point on the sky.

    Parameters
    ----------
    slope : `float`
        Sets the 'slope' of how fast the basis function
        value changes with hour angle.
    penalty_val : `float` or `np.nan`, optional
        Sets the value for the part of the sky which is
        not rising (hour angle between 0 and 180).
        Using a value of np.nan will mask this region of sky,
        a value of 0 will just make this non-rewarding.
    nside : `int` or None, optional
        Nside for the healpix map, default of None uses scheduler default.
    """

    def __init__(self, slope=0.1, penalty_val=0, nside=DEFAULT_NSIDE):
        super().__init__(nside=nside)
        self.slope = slope
        self.penalty_val = penalty_val

    # Probably not needed
    def check_feasibility(self, conditions):
        return True

    def _calc_value(self, conditions, indx=None):
        # HA should be available in the conditions object
        value = self.slope * conditions.HA
        past_zenith = np.where(conditions.HA < np.pi)
        value[past_zenith] = self.penalty_val
        return value



import unittest

import numpy as np

from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.scheduler.schedulers import ComCamBandSched, SimpleBandSched
from rubin_scheduler.utils import SURVEY_START_MJD


class TestBandSchedulers(unittest.TestCase):
    def test_ComCamBandSched(self):
        illum_bins = np.arange(0, 100 + 1, 50)
        band_groups = (("g", "r", "i"), ("i", "z", "y"))
        bandsched = ComCamBandSched(illum_bins=illum_bins, loaded_band_groups=band_groups)
        mjd = SURVEY_START_MJD
        conditions = Conditions(nside=8, mjd=mjd)
        conditions.moon_phase_sunset = 0
        load_bands = bandsched(conditions)
        self.assertTrue(load_bands == ["g", "r", "i"])
        conditions.moon_phase_sunset = 40
        load_bands = bandsched(conditions)
        self.assertTrue(load_bands == ["g", "r", "i"])
        conditions.moon_phase_sunset = 60
        load_bands = bandsched(conditions)
        self.assertTrue(load_bands == ["i", "z", "y"])
        conditions.moon_phase_sunset = 100
        load_bands = bandsched(conditions)
        self.assertTrue(load_bands == ["i", "z", "y"])

    def test_comcambandsched_except(self):
        illum_bins = np.arange(0, 100 + 1, 25)
        band_groups = (("g", "r", "i"), ("i", "z", "y"))
        with self.assertRaises(ValueError):
            ComCamBandSched(illum_bins=illum_bins, loaded_band_groups=band_groups)

    def test_SimpleBandSched(self):
        bandsched = SimpleBandSched(illum_limit=40)
        brightmoon_result = ["g", "r", "i", "z", "y"]
        newmoon_result = ["u", "g", "r", "i", "z"]
        mjd = SURVEY_START_MJD
        conditions = Conditions(nside=8, mjd=mjd)
        conditions.moon_phase_sunset = 0
        load_bands = bandsched(conditions)
        self.assertTrue(load_bands == newmoon_result)
        conditions.moon_phase_sunset = 50
        load_bands = bandsched(conditions)
        self.assertTrue(load_bands == brightmoon_result)


if __name__ == "__main__":
    unittest.main()



__all__ = (
    "IntRounded",
    "int_binned_stat",
    "smallest_signed_angle",
    "SchemaConverter",
    "HpInComcamFov",
    "HpInLsstFov",
    "hp_kd_tree",
    "match_hp_resolution",
    "TargetoO",
    "SimTargetooServer",
    "restore_scheduler",
    "warm_start",
    "gnomonic_project_toxy",
    "gnomonic_project_tosky",
    "raster_sort",
    "run_info_table",
    "inrange",
    "season_calc",
    "create_season_offset",
    "thetaphi2xyz",
    "xyz2thetaphi",
    "mean_azimuth",
    "wrap_ra_dec",
    "rotx",
)

import copy
import datetime
import os
import socket
import sqlite3
import warnings

import healpy as hp
import matplotlib.path as mplPath
import numpy as np
import pandas as pd

from rubin_scheduler import __version__
from rubin_scheduler.scheduler.utils.observation_array import ObservationArray
from rubin_scheduler.utils import (
    DEFAULT_NSIDE,
    _build_tree,
    _hpid2_ra_dec,
    _xyz_from_ra_dec,
    xyz_angular_radius,
)

rsVersion = __version__


def smallest_signed_angle(a1, a2):
    """
    via https://stackoverflow.com/questions/1878907/
    the-smallest-difference-between-2-angles
    """
    two_pi = 2.0 * np.pi
    x = a1 % two_pi
    y = a2 % two_pi
    a = (x - y) % two_pi
    b = (y - x) % two_pi
    result = b + 0
    alb = np.where(a < b)[0]
    result[alb] = -1.0 * a[alb]
    return result


def thetaphi2xyz(theta, phi):
    """Convert theta,phi to x,y,z position on the unit sphere

    Parameters
    ----------
    theta : `float`
        Theta coordinate in radians (should be 0-2pi).
    phi : `float`
        Phi coordinate in radians (should run from 0-pi).
    """
    x = np.sin(phi) * np.cos(theta)
    y = np.sin(phi) * np.sin(theta)
    z = np.cos(phi)
    return x, y, z


def xyz2thetaphi(x, y, z):
    """x,y,z position on unit sphere to theta,phi coords."""
    phi = np.arccos(z)
    theta = np.arctan2(y, x)
    return theta, phi


def rotx(theta, x, y, z):
    """rotate the x,y,z points theta radians about x axis"""
    sin_t = np.sin(theta)
    cos_t = np.cos(theta)
    xp = x
    yp = y * cos_t + z * sin_t
    zp = -y * sin_t + z * cos_t
    return xp, yp, zp


def wrap_ra_dec(ra, dec):
    # XXX--from MAF, should put in general utils
    """
    Wrap RA into 0-2pi and Dec into +/0 pi/2.

    Parameters
    ----------
    ra : numpy.ndarray
        RA in radians
    dec : numpy.ndarray
        Dec in radians

    Returns
    -------
    numpy.ndarray, numpy.ndarray
        Wrapped RA/Dec values, in radians.
    """
    # Wrap dec.
    low = np.where(dec < -np.pi / 2.0)[0]
    dec[low] = -1 * (np.pi + dec[low])
    ra[low] = ra[low] - np.pi
    high = np.where(dec > np.pi / 2.0)[0]
    dec[high] = np.pi - dec[high]
    ra[high] = ra[high] - np.pi
    # Wrap RA.
    ra = ra % (2.0 * np.pi)
    return ra, dec


def mean_azimuth(az, min_val=0.1):
    """Compute the mean azimuth value accounting for wrap

    Parameters
    ----------
    az : `array-like`
        The azimuths to average. Radians
    min_val : `float`
        A min cutoff to just use pi as the mean. Default 0.1. Radians
    """

    x = np.cos(az)
    y = np.sin(az)
    meanx = np.mean(x)
    meany = np.mean(y)
    angle = np.arctan2(meany, meanx)
    radius = np.sqrt(meanx**2 + meany**2)
    mid_az = angle % (2.0 * np.pi)
    if IntRounded(radius) < IntRounded(min_val):
        mid_az = np.pi
    return mid_az


class IntRounded:
    """
    Class to help force comparisons be made on scaled up integers,
    preventing machine precision issues cross-platforms.

    Danger zone would be doing comparisons with differences between
    `scale` and `limit`, e.g.:
    (1 + 1e-9) == (1 + 1e-10)
    returns False as expected, but
    IntRounded(1 + 1e-9) == intRounded(1 + 1e-10)
    will return True because it rounds off at 9 decimal places.

    Note that casting a NaN to an int can have different behaviours
    cross-platforms, so will throw an error if attempted.

    Parameters
    ----------
    inval : number-like thing
        Some number/array that we will want to compare
    scale : `float`
        How much to scale inval before rounding and converting to an int.
        Essentially how many digits of precesion to use in comparisons.
        Default 1e6.
    limit : `float`
        If values have a differce less than the limit, force comparisions
        to be done with scaled ints. Should be larger than whatever you
        think the machine floating point precision might be.
        Default 1e-9.
    """

    def __init__(self, inval, scale=1e6, limit=1e-9):
        if np.any(~np.isfinite(inval)):
            raise ValueError("IntRounded can only take finite values.")
        self.initial = inval
        self.scale = scale
        self.limit = limit

    def check_limit(self, v1, v2):
        """See if two values are potentiall close to floating precision limit.
        If they are close, return scaled up integers
        """
        v1 = np.atleast_1d(copy.copy(v1))
        v2 = np.atleast_1d(copy.copy(v2))

        non_zero_v1 = np.where(v1 != 0)
        non_zero_v2 = np.where(v2 != 0)

        # Don't need to worry if comparing arrays
        # that are empty or only zeros
        if (len(v1[non_zero_v1]) == 0) | (len(v2[non_zero_v2]) == 0):
            return v1, v2

        # Scale factor so we correctly compare things like 1e-25 to 1e-26
        # Kinda messy to avoid zeros or negatives to log.
        order_of_mag = np.min(
            [
                np.floor(np.log10(np.abs(v1[non_zero_v1]))).min(),
                np.floor(np.log10(np.abs(v2[non_zero_v2]))).min(),
            ]
        )
        scale_factor = 10**order_of_mag

        ab_diff = np.abs(v1 / scale_factor - v2 / scale_factor)
        close = np.where((ab_diff <= self.limit))[0]
        # There are some values with a potential round error
        if close.size > 0:
            floor = np.floor(np.minimum(v1, v2))
            v1 = np.round(v1 - floor * self.scale).astype(int)
            v2 = np.round(v2 - floor * self.scale).astype(int)
        return v1, v2

    def __eq__(self, other):
        v1, v2 = self.check_limit(self.initial, other.initial)
        return v1 == v2

    def __ne__(self, other):
        v1, v2 = self.check_limit(self.initial, other.initial)
        return v1 != v2

    def __lt__(self, other):
        v1, v2 = self.check_limit(self.initial, other.initial)
        return v1 < v2

    def __le__(self, other):
        v1, v2 = self.check_limit(self.initial, other.initial)
        return v1 <= v2

    def __gt__(self, other):
        v1, v2 = self.check_limit(self.initial, other.initial)
        return v1 > v2

    def __ge__(self, other):
        v1, v2 = self.check_limit(self.initial, other.initial)
        return v1 >= v2

    def __repr__(self):
        return str(self.initial)

    def __add__(self, other):
        out_scale = np.min([self.scale, other.scale])
        result = IntRounded(self.initial + other.initial, scale=out_scale)
        return result

    def __sub__(self, other):
        out_scale = np.min([self.scale, other.scale])
        result = IntRounded(self.initial - other.initial, scale=out_scale)
        return result

    def __mul__(self, other):
        out_scale = np.min([self.scale, other.scale])
        result = IntRounded(self.initial * other.initial, scale=out_scale)
        return result

    def __div__(self, other):
        out_scale = np.min([self.scale, other.scale])
        result = IntRounded(self.initial / other.initial, scale=out_scale)
        return result


def restore_scheduler(observation_id, scheduler, observatory, in_obs, band_sched=None, fast=True):
    """Put the scheduler and observatory in the state they were in.
    Handy for checking reward fucnction

    Parameters
    ----------
    observation_id : int
        The ID of the last observation that should be completed
    scheduler : rubin_scheduler.scheduler.scheduler object
        Scheduler object.
    observatory : rubin_scheduler.scheduler.observatory.Model_observatory
        The observaotry object
    in_obs : np.array or str
        Array of observations (formated like
        rubin_scheduler.scheduler.ObservationArray). If a string,
        assumed to be a file and SchemaConverter is used to load it.
    band_sched : rubin_scheduler.scheduler.scheduler object
        The band scheduler. Note that we don't look up the official
        end of the previous night, so there is potential for the
        loaded bands to not match.
    fast : bool (True)
        If True, loads observations and passes them as an array to
        the `add_observations_array` method. If False,
        passes observations individually with `add_observation` method.
    """
    if isinstance(in_obs, str):
        sc = SchemaConverter()
        # load up the observations
        observations = sc.opsim2obs(in_obs)
    else:
        observations = in_obs
    good_obs = np.where(observations["ID"] <= observation_id)[0]
    observations = observations[good_obs]

    if fast:
        scheduler.add_observations_array(observations)
        obs = observations[-1]
    else:
        for obs in observations:
            scheduler.add_observation(obs)

    if band_sched is not None:
        # We've assumed the band scheduler doesn't have any bands
        # May need to call the add_observation method on it if that
        # changes.

        # Make sure we have mounted the right bands for the night
        # XXX--note, this might not be exact, but should work most
        # of the time.
        mjd_start_night = np.min(observations["mjd"][np.where(observations["night"] == obs["night"])])
        observatory.mjd = mjd_start_night
        conditions = observatory.return_conditions()
        bands_needed = band_sched(conditions)
    else:
        bands_needed = ["u", "g", "r", "i", "y"]

    # update the observatory
    observatory.mjd = obs["mjd"] + observatory.observatory.visit_time(obs) / 3600.0 / 24.0
    observatory.obs_id_counter = obs["ID"] + 1
    observatory.observatory.parked = False
    observatory.observatory.current_ra_rad = obs["RA"]
    observatory.observatory.current_dec_rad = obs["dec"]
    observatory.observatory.current_rot_sky_pos_rad = obs["rotSkyPos"]
    observatory.observatory.cumulative_azimuth_rad = obs["cummTelAz"]
    observatory.observatory.current_band = obs["band"]
    observatory.observatory.mounted_bands = bands_needed
    # Note that we haven't updated last_az_rad, etc, but those
    # values should be ignored.

    return scheduler, observatory


def int_binned_stat(ids, values, statistic=np.mean):
    """
    Like scipy.binned_statistic, but for unique int ids
    """

    uids = np.unique(ids)
    order = np.argsort(ids)

    ordered_ids = ids[order]
    ordered_values = values[order]

    left = np.searchsorted(ordered_ids, uids, side="left")
    right = np.searchsorted(ordered_ids, uids, side="right")

    stat_results = []
    for le, ri in zip(left, right):
        stat_results.append(statistic(ordered_values[le:ri]))

    return uids, np.array(stat_results)


def gnomonic_project_toxy(ra1, dec1, r_acen, deccen):
    """Calculate x/y projection of ra1/dec1 in system with center
    at r_acen, deccen. Input radians. Grabbed from sims_selfcal"""
    # also used in Global Telescope Network website
    cosc = np.sin(deccen) * np.sin(dec1) + np.cos(deccen) * np.cos(dec1) * np.cos(ra1 - r_acen)
    x = np.cos(dec1) * np.sin(ra1 - r_acen) / cosc
    y = (np.cos(deccen) * np.sin(dec1) - np.sin(deccen) * np.cos(dec1) * np.cos(ra1 - r_acen)) / cosc
    return x, y


def gnomonic_project_tosky(x, y, r_acen, deccen):
    """Calculate RA/dec on sky of object with x/y and RA/Cen of
    field of view. Returns Ra/dec in radians."""
    denom = np.cos(deccen) - y * np.sin(deccen)
    RA = r_acen + np.arctan2(x, denom)
    dec = np.arctan2(np.sin(deccen) + y * np.cos(deccen), np.sqrt(x * x + denom * denom))
    return RA, dec


def match_hp_resolution(in_map, nside_out, unseen2nan=True):
    """Utility to convert healpix map resolution if needed and
    change hp.UNSEEN values to np.nan.

    Parameters
    ----------
    in_map : np.array
        A valie healpix map
    nside_out : int
        The desired resolution to convert in_map to
    unseen2nan : bool (True)
        If True, convert any hp.UNSEEN values to np.nan
    """
    current_nside = hp.npix2nside(np.size(in_map))
    if current_nside != nside_out:
        out_map = hp.ud_grade(in_map, nside_out=nside_out)
    else:
        out_map = in_map
    if unseen2nan:
        out_map[np.where(out_map == hp.UNSEEN)] = np.nan
    return out_map


def raster_sort(x0, order=["x", "y"], xbin=1.0):
    """XXXX--depriciated, use tsp instead.

    Do a sort to scan a grid up and down. Simple starting guess
    to traveling salesman.

    Parameters
    ----------
    x0 : array
    order : list
        Keys for the order x0 should be sorted in.
    xbin : float (1.)
        The bin_size to round off the first coordinate into

    returns
    -------
    array sorted so that it rasters up and down.
    """
    coords = x0.copy()
    bins = np.arange(
        coords[order[0]].min() - xbin / 2.0,
        coords[order[0]].max() + 3.0 * xbin / 2.0,
        xbin,
    )
    # digitize my bins
    coords[order[0]] = np.digitize(coords[order[0]], bins)
    order1 = np.argsort(coords, order=order)
    coords = coords[order1]
    places_to_invert = np.where(np.diff(coords[order[-1]]) < 0)[0]
    if np.size(places_to_invert) > 0:
        places_to_invert += 1
        indx = np.arange(coords.size)
        index_sorted = np.zeros(indx.size, dtype=int)
        index_sorted[0 : places_to_invert[0]] = indx[0 : places_to_invert[0]]

        for i, inv_pt in enumerate(places_to_invert[:-1]):
            if i % 2 == 0:
                index_sorted[inv_pt : places_to_invert[i + 1]] = indx[inv_pt : places_to_invert[i + 1]][::-1]
            else:
                index_sorted[inv_pt : places_to_invert[i + 1]] = indx[inv_pt : places_to_invert[i + 1]]

        if np.size(places_to_invert) % 2 != 0:
            index_sorted[places_to_invert[-1] :] = indx[places_to_invert[-1] :][::-1]
        else:
            index_sorted[places_to_invert[-1] :] = indx[places_to_invert[-1] :]
        return order1[index_sorted]
    else:
        return order1


class SchemaConverter:
    """
    Record how to convert an observation array to the standard
    opsim schema
    """

    def __init__(self):
        # Conversion dictionary, keys are opsim schema, values
        # are observation dtype names
        self.convert_dict = {
            "observationId": "ID",
            "night": "night",
            "observationStartMJD": "mjd",
            "observationStartLST": "lmst",
            "numExposures": "nexp",
            "visitTime": "visittime",
            "visitExposureTime": "exptime",
            "proposalId": "survey_id",
            "fieldId": "field_id",
            "fieldRA": "RA",
            "fieldDec": "dec",
            "altitude": "alt",
            "azimuth": "az",
            "band": "band",
            "filter": "filter",
            "airmass": "airmass",
            "skyBrightness": "skybrightness",
            "cloud": "clouds",
            "seeingFwhm500": "FWHM_500",
            "seeingFwhmGeom": "FWHM_geometric",
            "seeingFwhmEff": "FWHMeff",
            "fiveSigmaDepth": "fivesigmadepth",
            "slewTime": "slewtime",
            "slewDistance": "slewdist",
            "paraAngle": "pa",
            "pseudoParaAngle": "pseudo_pa",
            "rotTelPos": "rotTelPos",
            "rotTelPos_backup": "rotTelPos_backup",
            "rotSkyPos": "rotSkyPos",
            "rotSkyPos_desired": "rotSkyPos_desired",
            "moonRA": "moonRA",
            "moonDec": "moonDec",
            "moonAlt": "moonAlt",
            "moonAz": "moonAz",
            "moonDistance": "moonDist",
            "moonPhase": "moonPhase",
            "sunAlt": "sunAlt",
            "sunAz": "sunAz",
            "solarElong": "solarElong",
            "note": "note",
            "scheduler_note": "scheduler_note",
            "target_name": "target_name",
            "science_program": "science_program",
            "observation_reason": "observation_reason",
        }
        # For backwards compatibility
        self.backwards = {"target": "target_name"}
        # Column(s) not bothering to remap:
        # 'observationStartTime': None,
        self.inv_map = {v: k for k, v in self.convert_dict.items()}
        # angles to convert
        self.angles_rad2deg = [
            "fieldRA",
            "fieldDec",
            "altitude",
            "azimuth",
            "slewDistance",
            "paraAngle",
            "pseudoParaAngle",
            "rotTelPos",
            "rotSkyPos",
            "rotSkyPos_desired",
            "rotTelPos_backup",
            "moonRA",
            "moonDec",
            "moonAlt",
            "moonAz",
            "moonDistance",
            "sunAlt",
            "sunAz",
            "sunRA",
            "sunDec",
            "solarElong",
            "cummTelAz",
        ]
        # Put LMST into degrees too
        self.angles_hours2deg = ["observationStartLST"]

    def obs2opsim(self, obs_array, filename=None, info=None, delete_past=False, if_exists="append"):
        """Convert an array of observations into a pandas dataframe
        with Opsim schema.
        Parameters
        ----------
        obs_array : `np.array`
            Numpy array with OpSim observations.
        filename : `str`, optional
            Name of the database file to write to.
        info : `np.array`, optional
            Numpy array with database info.
        delete_past : `bool`
            Delete past observations (default=False)?
        if_exists : `str`
            Flag to pass to `to_sql` when writting to the
            database to control strategy when the database
            already exists.
        Returns
        -------
        `pd.DataFrame` or `None`
            Either the converted dataframe or `None`, if
            filename is provided.
        """
        if delete_past:
            try:
                os.remove(filename)
            except OSError:
                pass

        df = pd.DataFrame(obs_array)
        # TODO : Remove this hack which is for use with ts_scheduler
        # version <=v2.3 .. remove ts_scheduler actually drops "note".
        df.drop("note", axis=1, inplace=True)
        df = df.rename(index=str, columns=self.inv_map)
        for colname in self.angles_rad2deg:
            df[colname] = np.degrees(df[colname])
        for colname in self.angles_hours2deg:
            df[colname] = df[colname] * 360.0 / 24.0

        if filename is not None:
            con = sqlite3.connect(filename)
            df.to_sql("observations", con, index=False, if_exists=if_exists)
            if info is not None:
                df = pd.DataFrame(info)
                df.to_sql("info", con, if_exists=if_exists)
        else:
            return df

    def opsimdf2obs(self, df) -> np.recarray:
        """convert an opsim schema dataframe into an observation array.

        Parameters
        ----------
        df : `pd.DataFrame`
            Data frame containing opsim output observations.

        Returns
        -------
        obs : `np.recarray`
            Numpy array with OpSim observations.
        """
        # Do not modify the passed DataFrame, and avoid pandas getting
        # upset if a view is passed in.
        df = df.copy()

        # Make it backwards compatible if there are
        # columns that have changed names
        for key in self.backwards:
            if key in df.columns:
                df = df.rename(index=str, columns={key: self.backwards[key]})

        for key in self.angles_rad2deg:
            try:
                df[key] = np.radians(df[key])
            except (KeyError, TypeError):
                df[key] = np.nan
        for key in self.angles_hours2deg:
            try:
                df[key] = df[key] * 24.0 / 360.0
            except (KeyError, TypeError):
                df[key] = np.nan

        df = df.rename(index=str, columns=self.convert_dict)

        blank = ObservationArray()
        final_result = np.empty(df.shape[0], dtype=blank.dtype)
        # XXX-ugh, there has to be a better way.
        for key in final_result.dtype.names:
            if key in df.columns:
                final_result[key] = df[key].values
            else:
                warnings.warn(f"Column {key} not found.")

        return final_result

    def opsim2obs(self, filename):
        """convert an opsim schema dataframe into an observation array.

        Parameters
        ----------
        filename : `str`
            Sqlite file containing opsim output observations.
        """

        con = sqlite3.connect(filename)
        df = pd.read_sql("select * from observations;", con)
        return self.opsimdf2obs(df)


def hp_kd_tree(nside=DEFAULT_NSIDE, leafsize=100, scale=1e5):
    """
    Generate a KD-tree of healpixel locations

    Parameters
    ----------
    nside : int
        A valid healpix nside
    leafsize : int (100)
        Leafsize of the kdtree

    Returns
    -------
    tree : scipy kdtree
    """

    hpid = np.arange(hp.nside2npix(nside))
    ra, dec = _hpid2_ra_dec(nside, hpid)
    return _build_tree(ra, dec, leafsize, scale=scale)


class HpInLsstFov:
    """Return the healpixels in an underlying healpix grid that
    overlap an observation/pointing.

    This uses a very simple circular LSST camera model with no chip/raft gaps.

    Parameters
    ----------
    nside : `int`, optional
        Nside to match for the healpix array.
        Default None uses `set_default_nside`.
    fov_radius : `float`, optional
        Radius of the field of view in degrees. Default 1.75
        covers the inscribed circle.
    scale : `float`, optional
        How many sig figs to round when considering matches to healpixels.
        Useful for ensuring identical results cross-ploatform where
        float precision can vary.


    Examples
    --------
    Set up the class, then call to convert pointings to indices in the
    healpix array. Note that RA and dec should be in RADIANS.

    ```
    >>> ra = np.radians(30)
    >>> dec = np.radians(-20)
    >>> pointing2indx = HpInLsstFov()
    >>> indices = pointing2indx(ra, dec)
    >>> indices
    [8138, 8267]
    ```

    """

    def __init__(self, nside=DEFAULT_NSIDE, fov_radius=1.75, scale=1e5):
        self.tree = hp_kd_tree(nside=nside, scale=scale)
        self.radius = np.round(xyz_angular_radius(fov_radius) * scale).astype(int)
        self.scale = scale

    def __call__(self, ra, dec, **kwargs):
        """
        Parameters
        ----------
        ra : float, array
            RA in radians
        dec : float, array
            Dec in radians

        Returns
        -------
        indx : numpy array
            The healpixels that are within the FoV
        """
        x, y, z = _xyz_from_ra_dec(ra, dec)
        x = np.round(x * self.scale).astype(int)
        y = np.round(y * self.scale).astype(int)
        z = np.round(z * self.scale).astype(int)

        if np.size(x) == 1:
            indices = self.tree.query_ball_point((np.max(x), np.max(y), np.max(z)), self.radius)
        else:
            indices = self.tree.query_ball_point(np.vstack([x, y, z]).T, self.radius)
        return indices


class HpInComcamFov:
    """
    Return the healpixels within a ComCam pointing. Simple camera model
    with no chip gaps.
    """

    def __init__(self, nside=DEFAULT_NSIDE, side_length=0.7, scale=1e5):
        """
        Parameters
        ----------
        side_length : float (0.7)
            The length of one side of the square field of view (degrees).
        """
        self.nside = nside
        self.scale = scale
        self.tree = hp_kd_tree(nside=nside, scale=scale)
        self.side_length = np.round(np.radians(side_length * scale)).astype(int)
        self.inner_radius = np.round(xyz_angular_radius(side_length / 2.0) * scale).astype(int)
        self.outter_radius = np.round(xyz_angular_radius(side_length / 2.0 * np.sqrt(2.0)) * scale).astype(
            int
        )
        # The positions of the raft corners, unrotated
        self.corners_x = np.array(
            [
                -self.side_length / 2.0,
                -self.side_length / 2.0,
                self.side_length / 2.0,
                self.side_length / 2.0,
            ]
        )
        self.corners_y = np.array(
            [
                self.side_length / 2.0,
                -self.side_length / 2.0,
                -self.side_length / 2.0,
                self.side_length / 2.0,
            ]
        )

    def __call__(self, ra, dec, rotSkyPos=0.0):
        """
        Parameters
        ----------
        ra : float
            RA in radians
        dec : float
            Dec in radians
        rotSkyPos : float
            The rotation angle of the camera in radians
        Returns
        -------
        indx : numpy array
            The healpixels that are within the FoV
        """
        x, y, z = _xyz_from_ra_dec(ra, dec)
        x = np.round(x * self.scale).astype(int)
        y = np.round(y * self.scale).astype(int)
        z = np.round(z * self.scale).astype(int)
        # Healpixels within the inner circle
        indices = self.tree.query_ball_point((x, y, z), self.inner_radius)
        # Healpixels withing the outer circle
        indices_all = np.array(self.tree.query_ball_point((x, y, z), self.outter_radius))
        # Only need to check pixel if it is outside inner circle
        indices_to_check = indices_all[np.isin(indices_all, indices, invert=True)]

        if np.size(indices_to_check) == 0:
            ValueError("No HEALpix in pointing. Maybe need to increase nside.")

        cos_rot = np.cos(rotSkyPos)
        sin_rot = np.sin(rotSkyPos)
        x_rotated = self.corners_x * cos_rot - self.corners_y * sin_rot
        y_rotated = self.corners_x * sin_rot + self.corners_y * cos_rot

        # Draw the square that we want to check if points are in.
        bb_path = mplPath.Path(
            np.array(
                [
                    [x_rotated[0], y_rotated[0]],
                    [x_rotated[1], y_rotated[1]],
                    [x_rotated[2], y_rotated[2]],
                    [x_rotated[3], y_rotated[3]],
                    [x_rotated[0], y_rotated[0]],
                ]
            ).astype(int)
        )

        ra_to_check, dec_to_check = _hpid2_ra_dec(self.nside, indices_to_check)

        # Project the indices to check to the tangent plane, see
        # if they fall inside the polygon
        x, y = gnomonic_project_toxy(ra_to_check, dec_to_check, ra, dec)
        x = (x * self.scale).astype(int)
        y = (y * self.scale).astype(int)
        for i, xcheck in enumerate(x):
            # I wonder if I can do this all at once rather than a loop?
            if bb_path.contains_point((x[i], y[i])):
                indices.append(indices_to_check[i])

        return np.array(indices)


def run_info_table(observatory, extra_info=None):
    """
    Make a little table for recording the information about a run
    """

    observatory_info = observatory.get_info()
    if extra_info is not None:
        for key in extra_info:
            observatory_info.append([key, extra_info[key]])
    observatory_info = np.array(observatory_info)

    n_feature_entries = 3

    names = ["Parameter", "Value"]
    dtypes = ["|U200", "|U200"]
    result = np.zeros(observatory_info[:, 0].size + n_feature_entries, dtype=list(zip(names, dtypes)))

    # Fill in info about the run
    result[0]["Parameter"] = "Date, ymd"
    now = datetime.datetime.now()
    result[0]["Value"] = "%i, %i, %i" % (now.year, now.month, now.day)

    result[1]["Parameter"] = "hostname"
    result[1]["Value"] = socket.gethostname()

    result[2]["Parameter"] = "rubin_scheduler.__version__"
    result[2]["Value"] = rsVersion

    result[3:]["Parameter"] = observatory_info[:, 0]
    result[3:]["Value"] = observatory_info[:, 1]

    return result


def inrange(inval, minimum=-1.0, maximum=1.0):
    """
    Make sure values are within min/max
    """
    inval = np.array(inval)
    below = np.where(inval < minimum)
    inval[below] = minimum
    above = np.where(inval > maximum)
    inval[above] = maximum
    return inval


def warm_start(scheduler, observations, mjd_key="mjd"):
    """Replay a list of observations into the scheduler

    Parameters
    ----------
    scheduler : scheduler object
    observations : np.array
        An array of observation (e.g., from sqlite2observations)
    """

    # Check that observations are in order
    observations.sort(order=mjd_key)
    for observation in observations:
        scheduler.add_observation(observation)

    return scheduler


def season_calc(night, offset=0, modulo=None, max_season=None, season_length=365.25, floor=True):
    """
    Compute what season a night is in with possible offset and modulo
    using convention that night -365 to 0 is season -1.

    Parameters
    ----------
    night : int or array
        The night we want to convert to a season
    offset : float or array (0)
        Offset to be applied to night (days)
    modulo : int (None)
        If the season should be modulated (i.e., so we can get all
        even years) (seasons, years w/default season_length)
    max_season : int (None)
        For any season above this value (before modulo), set to -1
    season_length : float (365.25)
        How long to consider one season (nights)
    floor : bool (True)
        If true, take the floor of the season. Otherwise, returns
        season as a float
    """
    if np.size(night) == 1:
        night = np.ravel(np.array([night]))
    result = night + offset
    result = result / season_length
    if floor:
        result = np.floor(result)
    if max_season is not None:
        over_indx = np.where(IntRounded(result) >= IntRounded(max_season))

    if modulo is not None:
        neg = np.where(IntRounded(result) < IntRounded(0))
        result = result % modulo
        result[neg] = -1
    if max_season is not None:
        result[over_indx] = -1
    if floor:
        result = result.astype(int)
    return result


def create_season_offset(nside, sun_ra_rad):
    """
    Make an offset map so seasons roll properly
    """
    hpindx = np.arange(hp.nside2npix(nside))
    ra, dec = _hpid2_ra_dec(nside, hpindx)
    offset = ra - sun_ra_rad + 2.0 * np.pi
    offset = offset % (np.pi * 2)
    offset = offset * 365.25 / (np.pi * 2)
    offset = -offset - 365.25
    return offset


class TargetoO:
    """Class to hold information about a target of opportunity object

    Parameters
    ----------
    tooid : `int`
        Unique ID for the ToO.
    footprints : `np.array`
        np.array healpix maps. 1 for areas to observe, 0 for no observe.
        Can use np.nan for no-observe pixels, but that will be interpreted
        to mean the map cannot expand if the resolution chages.
    mjd_start : `float`
        The MJD the ToO starts
    duration : `float`
        Duration of the ToO (days).
    ra_rad_center : `float`
        RA of the estimated center of the event (radians).
    dec_rad_center : `float`
        Dec of the estimated center of the event (radians).
    too_type : `str`
        The type of ToO that is made.
    posterior_distance : `float`
        The posterior distance of the event. (kpc)
    """

    def __init__(
        self,
        tooid,
        footprint,
        mjd_start,
        duration,
        ra_rad_center=None,
        dec_rad_center=None,
        too_type=None,
        posterior_distance=None,
    ):
        self.footprint = footprint
        self.duration = duration
        self.id = tooid
        self.mjd_start = mjd_start
        self.ra_rad_center = ra_rad_center
        self.dec_rad_center = dec_rad_center
        self.too_type = too_type
        self.posterior_distance = posterior_distance


class SimTargetooServer:
    """Wrapper to deliver a targetoO object at the right time"""

    def __init__(self, targeto_o_list):
        self.targeto_o_list = targeto_o_list
        self.mjd_starts = np.array([too.mjd_start for too in self.targeto_o_list])
        durations = np.array([too.duration for too in self.targeto_o_list])
        self.mjd_ends = self.mjd_starts + durations

    def __call__(self, mjd):
        in_range = np.where((mjd > self.mjd_starts) & (mjd < self.mjd_ends))[0]
        result = None
        if in_range.size > 0:
            result = [self.targeto_o_list[i] for i in in_range]
        return result
